# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import uuid

import touchtechnology.common.db.models
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('competition', '0021_competition__data__person_uuid'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='clubassociation',
            unique_together=set([('club', 'person_uuid')]),
        ),
        migrations.AlterUniqueTogether(
            name='seasonassociation',
            unique_together=set([('season', 'person_uuid')]),
        ),
        migrations.AlterUniqueTogether(
            name='teamassociation',
            unique_together=set([('team', 'person_uuid')]),
        ),

        migrations.RemoveField(model_name='club', name='primary'),
        migrations.RemoveField(model_name='clubassociation', name='person'),
        migrations.RemoveField(model_name='seasonassociation', name='person'),
        migrations.RemoveField(model_name='teamassociation', name='person'),
        migrations.RemoveField(model_name='simplescorematchstatistic', name='player'),

        # Remove old related fields and rename new ones in place.
        migrations.RenameField('club', 'primary_uuid', 'primary'),
        migrations.RenameField('clubassociation', 'person_uuid', 'person'),
        migrations.RenameField('seasonassociation', 'person_uuid', 'person'),
        migrations.RenameField('teamassociation', 'person_uuid', 'person'),
        migrations.RenameField('simplescorematchstatistic', 'player_uuid', 'player'),

        # Modify the unique_together constraints again.
        migrations.AlterUniqueTogether(
            name='clubassociation',
            unique_together=set([('club', 'person')]),
        ),
        migrations.AlterUniqueTogether(
            name='seasonassociation',
            unique_together=set([('season', 'person')]),
        ),
        migrations.AlterUniqueTogether(
            name='teamassociation',
            unique_together=set([('team', 'person')]),
        ),

        # Remove the old primary pk and promote the new one.
        migrations.RemoveField(
            model_name='person',
            name='id',
        ),

        migrations.AlterField(
            model_name='person',
            name='uuid',
            field=models.UUIDField(
                default=uuid.uuid4,
                serialize=False,
                editable=False,
                unique=True,
            ),
        ),

        migrations.RunSQL(
            "SELECT 1;",
            "SELECT 1;",
            state_operations=[
                migrations.AlterField(
                    model_name='person',
                    name='uuid',
                    field=models.UUIDField(
                        primary_key=True,
                        default=uuid.uuid4,
                        serialize=False,
                        editable=False,
                    ),
                ),
            ],
        ),

        # Finally, resync with the models.py definitions
        migrations.AlterField(
            model_name='club',
            name='primary',
            field=touchtechnology.common.db.models.ForeignKey(
                to='competition.Person',
                related_name='+',
                blank=True,
                null=True,
                help_text='Appears on the front-end with other club information.',
                verbose_name='Primary contact',
            ),
        ),
        migrations.AlterField(
            model_name='clubassociation',
            name='person',
            field=touchtechnology.common.db.models.ForeignKey(to='competition.Person'),
        ),
        migrations.AlterField(
            model_name='seasonassociation',
            name='person',
            field=touchtechnology.common.db.models.ForeignKey(to='competition.Person'),
        ),
        migrations.AlterField(
            model_name='simplescorematchstatistic',
            name='player',
            field=touchtechnology.common.db.models.ForeignKey(related_name='statistics', to='competition.Person'),
        ),
        migrations.AlterField(
            model_name='teamassociation',
            name='person',
            field=touchtechnology.common.db.models.ForeignKey(to='competition.Person', null=True),
        ),
    ]



from django.test.utils import override_settings
from test_plus import TestCase


@override_settings(ROOT_URLCONF='example_app.urls')
class InvalidDateTest(TestCase):

    def test_archive_day(self):
        self.get('news:article', year='2013', month='feb', day='31')
        self.response_404()

    def test_article(self):
        self.get('news:article', year='2013', month='feb', day='31',
                 slug='tfms-new-generaltechnical-manager')
        self.response_404()


@override_settings(ROOT_URLCONF='example_app.urls')
class FeedTest(TestCase):

    def test_atom(self):
        self.assertGoodView('news:feed', format='atom')

    def test_rss(self):
        self.assertGoodView('news:feed', format='rss')



from __future__ import unicode_literals

import logging
import operator
import os
import re
import socket
from decimal import Decimal
from itertools import islice

import pkg_resources
import six
from classytags.arguments import Argument
from classytags.core import Options
from classytags.helpers import AsTag
from django.conf import settings
from django.db.models import Model, Q
from django.db.models.query import QuerySet
from django.forms.forms import BoundField
from django.forms.widgets import (
    CheckboxInput, CheckboxSelectMultiple, FileInput, MultiWidget, PasswordInput, RadioSelect,
    Select, Textarea, TextInput,
)
from django.template import Library, Node
from django.template.loader import get_template, render_to_string
from django.urls import Resolver404, resolve, reverse
from django.utils import timezone
from django.utils.encoding import smart_str
from django.utils.html import escape
from django.utils.http import urlencode
from django.utils.safestring import mark_safe
from django.utils.six.moves.urllib.parse import parse_qsl
from django.utils.text import slugify
from guardian.core import ObjectPermissionChecker
from namedentities import named_entities
from six.moves import xrange, zip_longest
from touchtechnology.common.default_settings import CURRENCY_SYMBOL
from touchtechnology.common.exceptions import NotModelManager
from touchtechnology.common.models import SitemapNode
from touchtechnology.common.utils import (
    create_exclude_filter, get_all_perms_for_model_cached, model_and_manager, tree_for_node,
)

logger = logging.getLogger(__name__)

camel_case_re = re.compile(
    r"(?P<new_word>((?<![A-Z])[A-Z]|[A-Z](?![A-Z0-9])))")

version_re = re.compile(r"""^
    (?P<version>\d+\.\d+)         # minimum 'N.N'
    (?P<extraversion>(?:\.\d+)*)  # any number of extra '.N' segments
    (?:
        (?P<prerel>[abc]|rc)         # 'a' = alpha, 'b' = beta
                                     # 'c' or 'rc' = release candidate
        (?P<prerelversion>\d+(?:\.\d+)*)
    )?
    (?P<postdev>(\.post(?P<post>\d+))?(\.dev(?P<dev>\d+))?)?
    $""", re.VERBOSE)

FORM_FIELD_TEMPLATE = get_template(
    "touchtechnology/common/templatetags/field.html")

register = Library()


@register.filter
def at_a_time(i, n):
    """
    Turn a list into smaller lists of n items. For example, if we have:

        object_list = [1, 2, 3, 4, 5, 6, 7, 8]

    The {% object_list|at_a_time:2 %} will result in:

        [[1, 2], [3, 4], [5, 6], [7, 8]

    Or {% object_list|at_a_time:3 %} will result in:

        [[1, 2, 3], [4, 5, 6], [7, 8, None]]

    We don't filter out None items, so check when iterating in your template.
    """
    return zip_longest(*[islice(i, x, None, n) for x in xrange(n)])


@register.filter
def camel_case_to_underscores(s):
    """Converts a CamelCase name to use lowercase_and_underscores.

    e.g. "ConvertHTMLToText" -> "convert_html_to_text"
    """
    def convert(match):
        content = match.group().lower()
        if match.start() == 0:
            return content
        else:
            return "_%s" % content
    return camel_case_re.sub(convert, s).lower()


@register.filter
def camel_case_split(s):
    """Converts a CamelCase name to be split by spaces.

    e.g. "CovertHTMLToText" -> "Convert HTML To Text"
    """
    def convert(match):
        content = match.group()
        if match.start() == 0:
            return content
        else:
            return " %s" % content
    return camel_case_re.sub(convert, s)


@register.filter
def cssify(s):
    return slugify(s or '').replace('-', '_')


@register.filter
def future(d):
    if d > timezone.now():
        return True
    return False


@register.filter
def split(st, on):
    return st.split(on)


@register.filter
def twittify(s):
    twitter_re = re.compile(r'(@(?P<username>[a-z0-9_]+)?)', re.I)
    return mark_safe(
        twitter_re.sub(r'@<a class="twitter user" '
                       'target="_blank" '
                       'href="http://twitter.com/\\2">\\2</a>', s))


@register.filter
def count(queryset):
    return queryset.count()


@register.filter
def disabled(queryset):
    return queryset.exclude(enabled=True)


@register.filter
def enabled(queryset):
    return queryset.filter(enabled=True)


@register.filter
def invisible(queryset):
    return queryset.exclude(hidden_from_navigation=False)


@register.filter
def visible(queryset):
    return queryset.filter(hidden_from_navigation=False)


@register.tag
def navigation(parser, token):
    """
    Build a navigational structure.

        Examples:

        {% navigation %}
        {% navigation current_node=node %}
        {% navigation start_at=1 stop_at=3 %}
        {% navigation root='page_3' %}
    """
    args = token.split_contents()[1:]
    kwargs = dict(map(lambda x: x.split('=', 1), args))
    kwargs = dict([(k, parser.compile_filter(v))
                   for k, v in kwargs.items()])
    return NavigationNode(kwargs)


class NavigationNode(Node):

    def __init__(self, kwargs):
        self.kwargs = kwargs

    def render(self, context):
        r_kwargs = dict([(str(k), v.resolve(context, ignore_failures=True))
                         for k, v in self.kwargs.items()])
        return do_navigation(**r_kwargs)


def do_navigation(root=None, start_at=None, stop_at=None, current_node=None,
                  expand_all_nodes=None, template_name=None, **kwargs):
    nodes = SitemapNode._tree_manager.select_related('content_type', 'parent')

    if template_name is None:
        template_name = 'touchtechnology/common/templatetags/navigation.html'

    if expand_all_nodes is None:
        expand_all_nodes = False

    logger.debug('======== do_navigation ========')
    logger.debug('template_name: %r', template_name)
    logger.debug('expand_all_nodes: %r', expand_all_nodes)
    logger.debug('current_node: %r', current_node)

    if root is not None:
        if isinstance(root, six.string_types):
            try:
                root = resolve(root).kwargs.get('node')
            except Resolver404:
                root = resolve(reverse(root)).kwargs.get('node')

        logger.debug('root: %r', root)

        try:
            nodes = root.get_descendants(include_self=True)
        except AttributeError:
            nodes = SitemapNode.objects.none()

        if current_node is not None and \
           current_node.tree_id == root.tree_id and \
           root.lft < current_node.lft and \
           not expand_all_nodes:
            nodes = tree_for_node(current_node)

    elif current_node is not None and not expand_all_nodes:
        nodes = tree_for_node(current_node)

    logger.debug('nodes: %r', nodes)

    # make sure we hide any nodes that are in a hidden part of the tree
    nodes_hidden_from_navigation = nodes.filter(
        Q(hidden_from_navigation=True) | Q(enabled=False))

    hidden_from_navigation = create_exclude_filter(
        nodes_hidden_from_navigation)

    nodes = nodes.exclude(hidden_from_navigation)

    logger.debug('nodes[cleaned]: %r', nodes)

    # flatten the list of nodes to a list
    tree = list(nodes)

    if current_node is None and not expand_all_nodes:
        stop_at = max(start_at or 0, stop_at or 0, 0)

    if start_at is not None:
        tree = [n for n in tree if n.level >= start_at]

    if stop_at is not None:
        tree = [n for n in tree if n.level <= stop_at]

    if not expand_all_nodes and current_node is not None:
        parents = []
        n = current_node
        while n.parent is not None:
            parents.append(n.parent)
            n = n.parent

        fmt = '[{rel}] {url} {node}'

        def func(node):
            """
            Filter function to determine if a node should appear in the
            navigation tree or not.
            """
            rel = current_node.rel(node)
            url = node.get_absolute_url()
            logger.debug(fmt.format(node=node, rel=rel, url=url))
            return rel in {'ROOT', 'ANCESTOR', 'PARENT', 'UNCLE',
                           'ME', 'SIBLING', 'DESCENDANT'}

        log = {
            'rel': 'NODE',
            'node': repr(current_node),
            'url': current_node.get_absolute_url(),
        }
        logger.debug(fmt.format(**log))
        tree = [t for t in tree if func(t)]

    # re-sort the queryset to get our correct tree structure back
    tree = sorted(tree, key=operator.attrgetter('tree_id', 'lft'))

    context = {
        'nodes': tree,
        'current_node': current_node,
        'hidden_nodes': nodes_hidden_from_navigation,
        'start_at': start_at,
        'stop_at': stop_at,
    }

    return render_to_string(template_name, context)


@register.simple_tag
def field(bf, label=None):
    if not isinstance(bf, BoundField):
        raise TypeError("{{% field %}} tag can only be used with "
                        "BoundFields ({0})".format(bf))

    if bf.is_hidden:
        return smart_str(bf)

    widget = bf.field.widget
    widget_class_name = camel_case_to_underscores(widget.__class__.__name__)

    if label is None:
        label = bf.label

    if isinstance(widget, (CheckboxInput,)):
        radio_checkbox_input = True
    else:
        radio_checkbox_input = False

    if label:
        if isinstance(widget, (TextInput, PasswordInput, FileInput,
                               Textarea, Select)) and not \
           isinstance(widget, (CheckboxInput, RadioSelect,
                               CheckboxSelectMultiple, MultiWidget)):
            # Use a <label> tag
            caption = bf.label_tag(label, attrs={'class': 'field_name'})
        else:
            # Don't use a <label> tag
            label_suffix = bf.form.label_suffix or ''
            caption = '<span class="field_name">' \
                      '%s%s</span>' % (label, label_suffix)
    else:
        caption = ''

    context = {
        'f': bf,
        'caption': caption,
        'widget_class_name': widget_class_name,
        'radio_checkbox_input': radio_checkbox_input,
        'no_label': not bool(label),
    }

    return FORM_FIELD_TEMPLATE.render(context)


@register.inclusion_tag('touchtechnology/common/templatetags/analytics.html')
def analytics(code=None):
    if code is None:
        code = getattr(settings, 'GOOGLE_ANALYTICS', code)
    context = {
        'code': code,
        'debug': not os.environ.get('SITE_ENV', 'dev') == 'live',
    }
    return context


@register.inclusion_tag('touchtechnology/common/templatetags/pagination.html',
                        takes_context=True)
def pagination(context):
    query_string = context.get('QUERY_STRING', '')
    query = parse_qsl(query_string, True)
    query = [q for q in query if q[0] != 'page']
    query_string = urlencode(query, True)
    context.update({'QUERY_STRING': query_string})
    return context


@register.filter('type')
def get_type(obj):
    if isinstance(obj, QuerySet):
        return obj.model._meta.verbose_name
    elif isinstance(obj, Model):
        return obj._meta.verbose_name
    else:
        return obj.__class__.__name__


@register.filter('types')
def get_type_plural(obj):
    if isinstance(obj, QuerySet):
        return obj.model._meta.verbose_name_plural
    elif isinstance(obj, Model):
        return obj._meta.verbose_name_plural
    else:
        # extremely naive
        return obj.__class__.__name__ + 's'


@register.filter
def htmlentities(s):
    replaced_entities = named_entities(
        escape(s).encode('ascii', 'xmlcharrefreplace').decode('utf8'))
    return mark_safe(replaced_entities)


@register.filter('abs')
def absolute_value(value):
    return abs(value)


@register.inclusion_tag('touchtechnology/common/templatetags/price.html')
def price(value, extra=''):
    context = {
        'SYMBOL': CURRENCY_SYMBOL,
        'AMOUNT': value or Decimal("0.00"),
        'EXTRA': extra,
    }
    return context


@register.filter('islice', is_safe=True)
def islice_(value, arg):
    """
    Returns an iterator slice of the list.
    """
    try:
        bits = []
        for x in arg.split(':'):
            if len(x) == 0:
                bits.append(None)
            else:
                bits.append(int(x))
        return islice(value, *bits)

    except (ValueError, TypeError):
        return value  # Fail silently.


@register.inclusion_tag('touchtechnology/common/templatetags/version.html')
def version(package, url=None):
    environment = pkg_resources.Environment()
    try:
        if package == 'python':
            context = {
                'name': "Python",
                'project': "Python",
                'version': environment.python,
                'release': None,
            }
        else:
            distribution = environment[package][-1]
            release = version_re.match(distribution.version).groupdict()
            context = {
                'name': distribution.key,
                'project': distribution.project_name,
                'version': distribution.version,
                'release': release.get('prerel'),
            }
    except IndexError:
        logger.exception('Attempting to lookup version for package "%s" not '
                         'in pkg_resources', package)
    if url:
        context['url'] = url
    return context


@register.filter('permchecker')
def create_permission_checker(model_or_manager, user):
    """
    Create an :class:`ObjectPermissionChecker` for optimal permission checking
    of related objects.

    http://django-guardian.readthedocs.io/en/stable/userguide/performance.html
    """
    model, manager = model_and_manager(model_or_manager)
    checker = ObjectPermissionChecker(user)
    checker.prefetch_perms(manager)
    return checker


@register.filter('checkperm')
def check_permission(obj, checker):
    """
    For a given model instance and a :class:`ObjectPermissionChecker` return
    which permissions the associated user has.
    """
    return [
        model_perm
        for model_perm in get_all_perms_for_model_cached(obj._meta.model)
        if checker.has_perm(model_perm, obj)
    ]


@register.filter('hasperm')
def has_permission(obj, user):
    try:
        model, manager = model_and_manager(obj)
    except NotModelManager:
        logger.exception('error="model cannot be determined"')
        return set()

    logger.debug('model="%s.%s", object_id="%s", user="%s"',
                 model._meta.app_label, model._meta.model_name,
                 getattr(obj, 'pk', ''), user.get_username())

    # Calculate the permissions this user has for the given object, both
    # directly and inferred by group memberships.
    perms = user.get_all_permissions(obj)
    perms |= user.get_group_permissions(obj)

    # What are the permissions that this model accepts?
    model_perms = get_all_perms_for_model_cached(model, ttl=300)

    # Superusers have all permissions, so add each permission that the model
    # accepts for this user to the set that has been explicitly cast.
    if user.is_superuser:
        perms.update(model_perms)

    # Otherwise we need to iterate the groups that the user belongs to and see
    # if they transfer permission to the user.
    else:
        for group in user.groups.all():
            group_perms = group.permissions.all()
            group_model_perms = [
                p.codename for p in model_perms if p in group_perms]
            logger.debug('user="%s", group="%s", permissions="%s"',
                         user.get_username(), group,
                         ", ".join(group_model_perms))
            perms.update(group_model_perms)

    # Log the resulting set of permissions for this user for this object.
    logger.debug('user="%s", permissions=%r', user.get_username(), perms)

    return perms


@register.simple_tag
def host():
    return socket.gethostname()


@register.inclusion_tag('touchtechnology/common/templatetags/hostname.html')
def hostname():
    hostname = socket.gethostname()
    host, domain = hostname.split('.', 1)
    return dict(hostname=hostname, host=host, domain=domain)


class Login(AsTag):
    options = Options(
        'as', Argument('varname', required=False, resolve=False))

    def get_value(self, context):
        request = context.get('request')
        url = reverse('accounts:login')
        if request:
            url += '?next=' + request.path
        return url


register.tag(Login)



from __future__ import unicode_literals

import datetime

import factory
import factory.fuzzy
from django.contrib.auth import get_user_model
from django.utils import timezone
from tournamentcontrol.competition import models, utils


class SitemapNodeBaseFactory(factory.DjangoModelFactory):
    """
    For any model which inherits from
    ``touchtechnology.common.models.SitemapNodeBase`` we need to execute the
    ``clean`` method which will populate the ``slug`` attribute according to
    the business logic of that method.
    """
    @factory.post_generation
    def post(obj, create, extracted, **kwargs):
        obj.clean()


class OrderedSitemapNodeFactory(SitemapNodeBaseFactory):
    """
    When adding a series of ``OrderedSitemapNode`` objects we should ensure
    that they are appropriately "ordered".
    """
    order = factory.Sequence(lambda n: (n + 1))


class RankDivisionFactory(OrderedSitemapNodeFactory):
    class Meta:
        model = models.RankDivision

    title = factory.Sequence(lambda n: "Division %d" % n)


class ClubFactory(SitemapNodeBaseFactory):
    class Meta:
        model = models.Club

    title = factory.Faker('company')


class CompetitionFactory(OrderedSitemapNodeFactory):
    class Meta:
        model = models.Competition

    title = factory.Sequence(lambda n: "Competition %d" % n)


class ClubRoleFactory(SitemapNodeBaseFactory):
    class Meta:
        model = models.ClubRole

    name = factory.Faker('company')
    competition = factory.SubFactory(CompetitionFactory)


class TeamRoleFactory(SitemapNodeBaseFactory):
    class Meta:
        model = models.TeamRole

    name = factory.Faker('company')
    competition = factory.SubFactory(CompetitionFactory)


class SeasonFactory(OrderedSitemapNodeFactory):
    class Meta:
        model = models.Season

    title = factory.Sequence(
        lambda n: str(range(2015, 1900, -1)[n]))
    timezone = factory.Faker('timezone')

    competition = factory.SubFactory(CompetitionFactory)


class SeasonExclusionDateFactory(factory.DjangoModelFactory):
    class Meta:
        model = models.SeasonExclusionDate

    date = factory.fuzzy.FuzzyDate(datetime.date(2013, 7, 15))
    season = factory.SubFactory(SeasonFactory)


class SeasonMatchTimeFactory(factory.DjangoModelFactory):
    class Meta:
        model = models.SeasonMatchTime

    season = factory.SubFactory(SeasonFactory)

    start = factory.LazyAttribute(lambda o: datetime.time(o.count))
    interval = factory.fuzzy.FuzzyChoice([25, 30, 35, 40])
    count = factory.fuzzy.FuzzyChoice(range(1, 15))


class VenueFactory(OrderedSitemapNodeFactory):
    class Meta:
        model = models.Venue

    title = factory.Sequence(lambda n: "Venue %d" % (n + 1))
    timezone = factory.SelfAttribute('season.timezone')

    season = factory.SubFactory(SeasonFactory)


class GroundFactory(OrderedSitemapNodeFactory):
    class Meta:
        model = models.Ground

    title = factory.Sequence(lambda n: "Ground %d" % (n + 1))
    timezone = factory.SelfAttribute('venue.timezone')

    venue = factory.SubFactory(VenueFactory)


class DivisionFactory(OrderedSitemapNodeFactory):
    class Meta:
        model = models.Division

    title = factory.Sequence(lambda n: "Division %d" % (n + 1))
    points_formula = "3*win + 2*draw + 1*loss"
    forfeit_for_score = 5
    forfeit_against_score = 0
    include_forfeits_in_played = True
    games_per_day = 2

    season = factory.SubFactory(SeasonFactory)


class DivisionExclusionDateFactory(factory.DjangoModelFactory):
    class Meta:
        model = models.DivisionExclusionDate

    date = factory.fuzzy.FuzzyDate(datetime.date(2013, 7, 15))
    division = factory.SubFactory(DivisionFactory)


class TeamFactory(OrderedSitemapNodeFactory):
    class Meta:
        model = models.Team

    title = factory.Sequence(lambda n: "Team %d" % (n + 1))

    club = factory.SubFactory(ClubFactory)
    division = factory.SubFactory(DivisionFactory)


class StageFactory(OrderedSitemapNodeFactory):
    class Meta:
        model = models.Stage

    title = factory.Sequence(lambda n: "Stage %d" % (n + 1))

    division = factory.SubFactory(DivisionFactory)


class StageGroupFactory(OrderedSitemapNodeFactory):
    class Meta:
        model = models.StageGroup

    title = factory.Sequence(lambda n: "Pool %d" % (n + 1))

    stage = factory.SubFactory(StageFactory)


class UndecidedTeamFactory(factory.DjangoModelFactory):
    class Meta:
        model = models.UndecidedTeam

    label = factory.Faker('text', max_nb_chars=30)
    stage = factory.SubFactory(StageFactory)


class MatchFactory(factory.DjangoModelFactory):
    class Meta:
        model = models.Match

    stage = factory.SubFactory(StageFactory)

    home_team = factory.SubFactory(
        TeamFactory, division=factory.SelfAttribute('..stage.division'))
    away_team = factory.SubFactory(
        TeamFactory, division=factory.SelfAttribute('..stage.division'))

    datetime = factory.fuzzy.FuzzyDateTime(
        datetime.datetime(2008, 1, 1, tzinfo=timezone.utc))

    date = factory.LazyAttribute(lambda o: o.datetime.date())
    time = factory.LazyAttribute(lambda o: o.datetime.time())


class UserFactory(factory.DjangoModelFactory):
    class Meta:
        model = get_user_model()

    first_name = factory.Faker('first_name')
    last_name = factory.Faker('last_name')

    username = factory.Sequence(lambda n: "user%04d" % (n + 1))
    email = factory.Sequence(lambda n: "user%03d@example.com" % (n + 1))

    password = 'crypt$$azoV33FJ2h3AA'  # pre-crypted "password"
    is_active = True


class PersonFactory(factory.DjangoModelFactory):
    class Meta:
        model = models.Person

    first_name = factory.SelfAttribute('user.first_name')
    last_name = factory.SelfAttribute('user.last_name')

    date_of_birth = factory.Faker(
        'date_time_between', start_date="-30y", end_date="-20y")
    gender = factory.fuzzy.FuzzyChoice(['M', 'F', 'X'])

    club = factory.SubFactory(ClubFactory)
    user = factory.SubFactory(UserFactory)


class ClubAssociationFactory(factory.DjangoModelFactory):
    class Meta:
        model = models.ClubAssociation

    club = factory.SubFactory(ClubFactory)
    person = factory.SubFactory(PersonFactory)


class TeamAssociationFactory(factory.DjangoModelFactory):
    class Meta:
        model = models.TeamAssociation

    team = factory.SubFactory(TeamFactory)
    person = factory.SubFactory(PersonFactory)


class DrawFormatFactory(factory.DjangoModelFactory):
    class Meta:
        model = models.DrawFormat

    teams = factory.fuzzy.FuzzyChoice([4, 6, 8, 12])
    name = factory.LazyAttribute(
        lambda a: 'Round Robin ({}/{} teams)'.format(a.teams - 1, a.teams))
    text = factory.LazyAttribute(
        lambda a: utils.round_robin_format([t for t in range(1, a.teams + 1)]))



from test_plus import TestCase
from touchtechnology.common.tests.factories import UserFactory
from tournamentcontrol.competition.models import Match
from tournamentcontrol.competition.tests import factories


class DrawGenerationMatchFormSetTest(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.superuser = UserFactory.create(is_staff=True, is_superuser=True)

        factories.DrawFormatFactory.reset_sequence()
        cls.draw_format = factories.DrawFormatFactory.create(teams=4)

        factories.StageFactory.reset_sequence()
        cls.stage = factories.StageFactory.create()

        factories.TeamFactory.reset_sequence()
        cls.teams = factories.TeamFactory.create_batch(
            4, division=cls.stage.division)

    @classmethod
    def tearDownClass(cls):
        pass

    def test_draw_format_name(self):
        self.assertEqual(self.draw_format.name, "Round Robin (3/4 teams)")

    def test_draw_format_text(self):
        self.assertEqual(
            self.draw_format.text,
            "ROUND\n"
            "1: 1 vs 4\n"
            "2: 2 vs 3\n"
            "ROUND\n"
            "3: 1 vs 3\n"
            "4: 4 vs 2\n"
            "ROUND\n"
            "5: 1 vs 2\n"
            "6: 3 vs 4"
        )

    def test_draw_generation_match_form_set_save(self):
        self.assertEqual(Match.objects.count(), 0)

        data0 = {
            'draw_generation_wizard-current_step': '0',

            '0-TOTAL_FORMS': '1',
            '0-INITIAL_FORMS': '1',
            '0-MIN_NUM_FORMS': '0',
            '0-MAX_NUM_FORMS': '1000',

            '0-0-start_date_0': '8',
            '0-0-start_date_1': '8',
            '0-0-start_date_2': '2018',
            '0-0-format': str(self.draw_format.pk),
            '0-0-rounds': '',
            '0-0-offset': '',
        }

        with self.login(self.superuser):
            self.post(
                'admin:fixja:competition:season:division:stage:draw:build',
                self.stage.division.season.competition.pk,
                self.stage.division.season.pk,
                self.stage.division.pk,
                self.stage.pk,
                data=data0)
            self.response_200()

            wizard = self.get_context('wizard')
            forms = wizard['form']
            total_forms = initial_forms = len(forms)

            data1 = {
                'draw_generation_wizard-current_step': '1',

                '1-TOTAL_FORMS': str(total_forms),
                '1-INITIAL_FORMS': str(initial_forms),
                '1-MIN_NUM_FORMS': '0',
                '1-MAX_NUM_FORMS': '1000',
            }

            for i, form in enumerate(forms):
                data1.update({
                    '1-%d-round' % i: str(form['round'].initial),
                    '1-%d-home_team' % i: str(form['home_team'].initial or ''),
                    '1-%d-away_team' % i: str(form['away_team'].initial or ''),
                    '1-%d-date' % i: form['date'].initial.strftime("%Y-%m-%d"),
                })

            self.post(
                'admin:fixja:competition:season:division:stage:draw:build',
                self.stage.division.season.competition.pk,
                self.stage.division.season.pk,
                self.stage.division.pk,
                self.stage.pk,
                data=data1)
            self.response_302()

        self.assertEqual(Match.objects.count(), 6)

        team1, team2, team3, team4 = self.teams
        self.assertQuerysetEqual(
            Match.objects.all(),
            [
                '<Match: 1: %s vs %s>' % (team1, team4),
                '<Match: 1: %s vs %s>' % (team2, team3),
                '<Match: 2: %s vs %s>' % (team1, team3),
                '<Match: 2: %s vs %s>' % (team4, team2),
                '<Match: 3: %s vs %s>' % (team1, team2),
                '<Match: 3: %s vs %s>' % (team3, team4),
            ],
        )



import imp
import logging
import os.path

from django.conf import settings
from django.conf.urls import include, url
from django.conf.urls.static import static
from django.contrib.sitemaps.views import sitemap
from django.contrib.staticfiles.urls import staticfiles_urlpatterns
from django.core.cache import cache
from django.core.exceptions import ObjectDoesNotExist, PermissionDenied
from django.core.files.storage import default_storage
from django.http import HttpResponseForbidden
from django.shortcuts import redirect, render_to_response
from django.template import RequestContext, TemplateDoesNotExist
from django.urls import reverse_lazy
from django.utils.deprecation import MiddlewareMixin
from django.utils.six.moves.urllib.parse import urlunparse
from guardian.conf import settings as guardian_settings
from mptt.utils import tree_item_iterator
from touchtechnology.common.default_settings import (
    SITEMAP_HTTPS_OPTION, SITEMAP_ROOT,
)
from touchtechnology.common.models import SitemapNode
from touchtechnology.common.sitemaps import NodeSitemap
from touchtechnology.common.sites import AccountsSite
from touchtechnology.common.views import login
from touchtechnology.content.models import Redirect
from touchtechnology.content.views import dispatch

DEHYDRATED_URLPATTERNS_KEY = 'urlpatterns'
DEHYDRATED_URLPATTERNS_TIMEOUT = 600

logger = logging.getLogger(__name__)
protect = AccountsSite(name="protect")


class SitemapNodeMiddleware(MiddlewareMixin):

    def process_request(self, request):
        # When working with multiple tenants, we want to shard the cache for
        # each of them. Use of the version is a nice way to do this as it will
        # prevent collisions while making the API consistent.
        v_kw = {}
        if hasattr(request, 'tenant'):
            v_kw.setdefault('version', request.tenant.schema_name)

        # TODO write a cache backend that will do this automatically, and
        #      contribute it back to django-tenant-schemas

        dehydrated = cache.get(DEHYDRATED_URLPATTERNS_KEY, [], **v_kw)

        if not dehydrated:
            logging.getLogger('newrelic.cache').debug('RECALCULATE URLCONF')

            # We need a secret set of account urls so we can bounce the user
            # here if the page is protected. As we haven't yet embedded our
            # url conf (we're in the middle of building it!) this will need
            # to be found using a reverse_lazy below.
            try:
                root = SitemapNode._tree_manager.root_nodes().first()
            except ObjectDoesNotExist:
                root = None
            dehydrated.append({
                'regex': r'^p/',
                'site': protect,
                'kwargs': {
                    'node': root,
                },
            })

            enabled_nodes = SitemapNode._tree_manager.all()
            related_nodes = enabled_nodes.select_related('content_type')

            def has_disabled_ancestors(st):
                for ancestor in st['ancestors']:
                    if not ancestor.enabled:
                        return True
                return False

            def get_absolute_url(n, st):
                assert not n.is_root_node()
                offset = 1 if st['ancestors'][0].slug == SITEMAP_ROOT else 0
                paths = [
                    ancestor.slug
                    for ancestor in st['ancestors'][offset:]
                ]
                if paths:
                    return os.path.join(os.path.join(*paths), n.slug)
                return n.slug

            for node, struct in tree_item_iterator(related_nodes, True, lambda x: x):
                # Skip over nodes that they themselves or have disabled ancestors.
                if not node.enabled:
                    logger.debug('%r is disabled, omit from urlconf', node)
                    continue
                if has_disabled_ancestors(struct):
                    logger.debug('%r has disabled ancestor, omit from urlconf', node)
                    continue

                if node.is_root_node() and node.slug == SITEMAP_ROOT:
                    path = ''
                elif node.is_root_node():
                    path = node.slug
                else:
                    path = get_absolute_url(node, struct)

                if path and settings.APPEND_SLASH:
                    path += '/'

                if node.content_type is not None and node.content_type.model == 'placeholder':
                    try:
                        app = node.object.site(node)
                    except (AttributeError, ImportError, ValueError):
                        logger.exception("Application is unavailable, "
                                         "disabling this node.")
                        node.disable()
                    else:
                        pattern = {
                            'regex': r'^%s' % path,
                            'site': app,
                            'kwargs': dict(node=node, **node.kwargs),
                            'name': app.name,
                        }
                        # When nesting applications we need to ensure that any
                        # root url is not clobbered by the patterns of the
                        # parent application. In these cases, force them to the
                        # top of the map.
                        if node.parent and node.parent.content_type and \
                                node.parent.content_type.model == 'placeholder':
                            dehydrated.insert(0, pattern)
                        else:
                            dehydrated.append(pattern)

                elif node.object_id is None:
                    dehydrated.append({
                        'regex': r'^%s$' % path,
                        'view': dispatch,
                        'kwargs': dict(
                            node=node, url=path),
                        'name': 'folder_%d' % node.pk,
                    })

                else:
                    dehydrated.append({
                        'regex': r'^%s$' % path,
                        'view': dispatch,
                        'kwargs': dict(
                            page_id=node.object_id, node=node, url=path),
                        'name': 'page_%d' % (
                            node.object_id if node.object_id else None,),
                    })

            cache.set(DEHYDRATED_URLPATTERNS_KEY, dehydrated,
                      timeout=DEHYDRATED_URLPATTERNS_TIMEOUT,
                      **v_kw)

        # Always start with the project wide ROOT_URLCONF and add our
        # sitemap.xml view
        urlpatterns = [
            url(r'^', include(settings.ROOT_URLCONF)),
            url(r'^sitemap\.xml', sitemap,
                {'sitemaps': {'nodes': NodeSitemap}}, name='sitemap'),
        ]

        # Construct the cache of url pattern definitions. We are not keeping
        # the actual patterns, because pickling is problematic for the .url
        # instancemethod - instead we keep the skeleton and build it on the
        # fly from cache... rehydrating it ;)
        for node in dehydrated:
            try:
                pattern = url(node['regex'], node['view'],
                              node['kwargs'], name=node.get('name'))
            except KeyError:
                pattern = url(node['regex'], node['site'].urls,
                              node['kwargs'], name=node['site'].name)
            urlpatterns.append(pattern)

        # For development, add the MEDIA_URL and STATIC_URL to the project
        if settings.DEBUG:
            urlpatterns += static(
                settings.MEDIA_URL,
                document_root=default_storage.path(''),
                show_indexes=True)

            urlpatterns += staticfiles_urlpatterns()

        # Create a new module on the fly and attach the rehydrated urlpatterns
        dynamic_urls = imp.new_module('dynamic_urls')
        dynamic_urls.urlpatterns = urlpatterns

        # Attach the module to the request
        request.urlconf = dynamic_urls

    def process_view(self, request, view_func, view_args, view_kwargs):
        node = view_kwargs.get('node')

        if node:
            required = node.restrict_to_groups.all()

            if required:
                next = node.get_absolute_url()
                to = reverse_lazy('accounts:login')

                # An anonymous user will never be a member of a group, so make
                # them go off and be authenticated.
                if request.user.is_anonymous():
                    return login(request, to=to, next=next)

                # A user who is not a member of a suitable group should get a
                # 403 page.
                groups = request.user.groups.all()
                if set(required).difference(groups):
                    if guardian_settings.RENDER_403:
                        try:
                            response = render_to_response(
                                guardian_settings.TEMPLATE_403, {},
                                RequestContext(request))
                            response.status_code = 403
                            return response
                        except TemplateDoesNotExist as e:
                            if settings.DEBUG:
                                raise e
                    elif guardian_settings.RAISE_403:
                        raise PermissionDenied
                    return HttpResponseForbidden()

        if SITEMAP_HTTPS_OPTION and node and \
           node.require_https and not request.is_secure():
            host = request.META.get('HTTP_HOST')
            path = request.META.get('PATH_INFO')
            redirect_to = urlunparse(
                ('https', host, path, '', '', ''))
            return redirect(redirect_to)


def redirect_middleware(get_response):
    def middleware(request):
        try:
            obj = Redirect.objects.get(source_url__exact=request.path)
        except Redirect.DoesNotExist:
            response = get_response(request)
        else:
            response = redirect(obj.destination_url, obj.permanent)
        return response
    return middleware



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import tournamentcontrol.competition.models
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('competition', '0016_season_forfeit_notifications'),
    ]

    operations = [
        migrations.AddField(
            model_name='competition',
            name='rank_importance',
            field=models.DecimalField(max_digits=6, decimal_places=3, blank=True, null=True),
        ),
        migrations.AddField(
            model_name='season',
            name='rank_importance',
            field=models.DecimalField(max_digits=6, decimal_places=3, blank=True, null=True),
        ),
        migrations.AddField(
            model_name='division',
            name='rank_importance',
            field=models.DecimalField(max_digits=6, decimal_places=3, blank=True, null=True),
        ),
        migrations.AddField(
            model_name='stage',
            name='rank_importance',
            field=models.DecimalField(max_digits=6, decimal_places=3, blank=True, null=True),
        ),
        migrations.AddField(
            model_name='stagegroup',
            name='rank_importance',
            field=models.DecimalField(max_digits=6, decimal_places=3, blank=True, null=True),
        ),
        migrations.AddField(
            model_name='match',
            name='rank_importance',
            field=models.DecimalField(max_digits=6, decimal_places=3, blank=True, null=True),
        ),
        migrations.CreateModel(
            name='RankDivision',
            fields=[
                ('id', models.AutoField(verbose_name='ID', auto_created=True, primary_key=True, serialize=False)),
                ('title', models.CharField(verbose_name='Title', max_length=255)),
                ('short_title', models.CharField(verbose_name='Short title', blank=True, help_text='This is used in navigation menus instead of the longer title value.', max_length=100)),
                ('slug', models.SlugField(verbose_name='Slug', max_length=255)),
                ('slug_locked', models.BooleanField(verbose_name='Slug locked', default=False)),
                ('order', models.PositiveIntegerField(default=1)),
                ('enabled', models.BooleanField(default=True)),
            ],
            options={
                'ordering': ('order',),
                'abstract': False,
            },
            bases=(tournamentcontrol.competition.models.AdminUrlMixin, models.Model),
        ),
        migrations.CreateModel(
            name='RankPoints',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('points', models.DecimalField(default=0, max_digits=6, decimal_places=3)),
                ('date', models.DateField()),
            ],
            options={
                'ordering': ('-date', 'team', '-points'),
            },
        ),
        migrations.CreateModel(
            name='RankTeam',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
            options={
                'ordering': ('division',),
            },
        ),
        migrations.AddField(
            model_name='division',
            name='rank_division',
            field=models.ForeignKey(null=True, blank=True, to='competition.RankDivision'),
        ),
        migrations.AddField(
            model_name='team',
            name='rank_division',
            field=models.ForeignKey(null=True, blank=True, to='competition.RankDivision'),
        ),
        migrations.AddField(
            model_name='rankteam',
            name='club',
            field=models.ForeignKey(to='competition.Club'),
        ),
        migrations.AddField(
            model_name='rankteam',
            name='division',
            field=models.ForeignKey(to='competition.RankDivision'),
        ),
        migrations.AlterUniqueTogether(
            name='rankteam',
            unique_together=set([('club', 'division')]),
        ),
        migrations.AddField(
            model_name='rankpoints',
            name='team',
            field=models.ForeignKey(to='competition.RankTeam'),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import datetime

import touchtechnology.common.db.models
from django.db import migrations
from django.utils.timezone import utc


class Migration(migrations.Migration):

    dependencies = [
        ('news', '0002_article_last_modified'),
    ]

    operations = [
        migrations.AddField(
            model_name='category',
            name='last_modified',
            field=touchtechnology.common.db.models.DateTimeField(default=datetime.datetime(2015, 3, 23, 11, 11, 17, 201480, tzinfo=utc), auto_now=True),
            preserve_default=False,
        ),
    ]



from django.apps import AppConfig, apps
from django.conf import settings
from django.core import checks
from django.utils.module_loading import import_string
from first import first


class AdminConfig(AppConfig):
    name = 'touchtechnology.admin'
    label = 'touchtechnology_admin'


@checks.register('config')
def touchtechnology_assumptions(app_configs, **kwargs):
    """
    When building a touchtechnology.admin driven piece of functionality, we
    make plenty of opinionated assumptions about what will be happening. So we
    don't shoot ourselves in the foot too often, these checks should be run to
    ensure we are following our conventions.
    """
    if app_configs is None:
        app_configs = apps.get_app_configs()

    INSTALLED_APPS = [
        config.name
        for config in app_configs
    ]

    errors = []

    # determine if we are running a multi-tenanted installation or not
    MULTITENANT = 'tenant_schemas' in INSTALLED_APPS
    SITES = 'django.contrib.sites' in INSTALLED_APPS

    if not MULTITENANT and not SITES:
        errors.append(checks.Critical(
            "touchtechnology.admin requires either 'django-tenant-schemas' or "
            "'django.contrib.sites' to be in INSTALLED_APPS",
            hint="Add either 'tenant_schemas' or 'django.contrib.sites' to "
                 "INSTALLED_APPS",
        ))

    # ensure appropriate tenant/site middleware is first loaded
    try:
        first_middleware = first(settings.MIDDLEWARE)
    except (AttributeError, TypeError):
        first_middleware = first(settings.MIDDLEWARE_CLASSES)

    if MULTITENANT:
        TenantMiddleware = import_string(
            'tenant_schemas.middleware.TenantMiddleware')
        FirstMiddleware = import_string(first_middleware)
        if not issubclass(FirstMiddleware, TenantMiddleware):
            errors.append(checks.Critical(
                "touchtechnology.admin requires a 'django-tenant-schemas' "
                "middleware to be listed first.",
                hint=None,
            ))
    elif SITES:
        if first_middleware not in (
                'django.contrib.sites.middleware.CurrentSiteMiddleware',):
            errors.append(checks.Critical(
                "touchtechnology.admin requires "
                "'django.contrib.sites.middleware.CurrentSiteMiddleware' to "
                "be listed first.",
                hint=None,
            ))
    else:
        errors.append(checks.Warning(
            "Make sure the appropriate middleware is listed first.",
            hint=None,
        ))

    REQUIRED_APPS = {
        'bootstrap3': 'django-bootstrap3',
        'django_gravatar': 'django-gravatar2',
    }

    for app, package in REQUIRED_APPS.items():
        if app not in INSTALLED_APPS:
            errors.append(checks.Critical(
                "touchtechnology.admin requires {}".format(package),
                hint='Add "{}" to INSTALLED_APPS'.format(app),
            ))

    return errors



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import tournamentcontrol.competition.models
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('competition', '0010_remove_match_video_approved_by'),
    ]

    operations = [
        migrations.AlterField(
            model_name='club',
            name='twitter',
            field=tournamentcontrol.competition.models.TwitterField(
                help_text='Official Twitter name for use in social "mentions"',
                max_length=50,
                blank=True,
            ),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='division',
            name='points_formula',
            field=tournamentcontrol.competition.models.LadderPointsField(
                null=True,
                verbose_name='Points system',
                blank=True,
            ),
            preserve_default=True,
        ),
    ]



from __future__ import unicode_literals

import base64
from datetime import timedelta
from operator import or_

from django.conf import settings
from django.conf.urls import include, url
from django.contrib import messages
from django.contrib.sitemaps import views as sitemaps_views
from django.db.models import Case, Count, F, Q, Sum, When
from django.http import Http404, HttpResponse, HttpResponseGone
from django.utils import timezone
from django.utils.encoding import smart_bytes
from django.utils.module_loading import import_string
from django.utils.translation import ugettext, ugettext_lazy as _
from icalendar import Calendar, Event
from six.moves import reduce
from touchtechnology.common.decorators import login_required_m
from touchtechnology.common.sites import Application
from tournamentcontrol.competition import rank
from tournamentcontrol.competition.decorators import competition_slug
from tournamentcontrol.competition.forms import (
    ConfigurationForm, MultiConfigurationForm, RankingConfigurationForm,
)
from tournamentcontrol.competition.models import Competition, Person


class CompetitionSite(Application):

    kwargs_form_class = ConfigurationForm

    def __init__(self, name='competition', app_name='competition', **kwargs):
        # store the node for future reference
        self.node = kwargs.get('node')
        super(CompetitionSite, self).__init__(
            name=name, app_name=app_name, **kwargs)
        self._competitions = Competition.objects.filter(enabled=True)
        if 'competition' in kwargs:
            self._competitions = self._competitions.filter(
                slug=kwargs['competition'])

    def season_urls(self):
        return [
            url(r'^$', self.season, name='season'),
            url(r'^forfeit/$', self.forfeit_list, name='forfeit-list'),
            url(r'^forfeit/(?P<match>\d+)/$', self.forfeit, name='forfeit'),
            url(r'^videos/$', self.season_videos, name='season-videos'),
            url(r'^club:(?P<club>[\w-]+)/$', self.club, name='club'),
            url(r'^club:(?P<club>[\w-]+).ics$', self.calendar, name='calendar'),
            url(r'^(?P<division>[\w-]+).ics$', self.calendar, name='calendar'),
            url(r'^(?P<division>[\w-]+)/$', self.division, name='division'),
            url(r'^(?P<division>[\w-]+):(?P<stage>[\w-]+)/$', self.stage, name='stage'),
            url(r'^(?P<division>[\w-]+):(?P<stage>[\w-]+):(?P<pool>[\w-]+)/$', self.pool, name='pool'),
            url(r'^(?P<division>[\w-]+)/match:(?P<match>\d+)/$', self.match, name='match'),
            url(r'^(?P<division>[\w-]+)/match:(?P<match>\d+)/video/$', self.match_video, name='match-video'),
            url(r'^(?P<division>[\w-]+)/(?P<team>[\w-]+).ics$', self.calendar, name='calendar'),
            url(r'^(?P<division>[\w-]+)/(?P<team>[\w-]+)/$', self.team, name='team'),
        ]

    def competition_urls(self):
        return [
            url(r'^$', self.competition, name='competition'),
            url(r'^(?P<season>[\w-]+).ics$', self.calendar, name='calendar'),
            url(r'^(?P<season>[\w-]+)/', include(self.season_urls())),
        ]

    def get_urls(self):
        if 'season' in self.kwargs:
            urlpatterns = [
                url(r'^', include(self.season_urls()),
                    kwargs=self.kwargs),
            ]
        elif 'competition' in self.kwargs:
            urlpatterns = [
                url(r'^', include(self.competition_urls()),
                    kwargs=self.kwargs),
            ]
        else:
            urlpatterns = [
                url(r'^$', self.index, name='index'),
                url(r'^(?P<competition>[\w-]+)/',
                    include(self.competition_urls())),
            ]
        return urlpatterns

    def sitemap_index(self, request, node=None, competition=None, season=None, division=None, *args, **kwargs):
        # We need to "pop" the node keyword argument
        return sitemaps_views.index(request, *args, **kwargs)

    def sitemap_section(self, request, node=None, competition=None, season=None, division=None, *args, **kwargs):
        # We need to "pop" the node keyword argument
        return sitemaps_views.sitemap(request, *args, **kwargs)

    @property
    def urls(self):
        return self.get_urls(), self.app_name, self.name

    @property
    def competitions(self):
        return self._competitions.prefetch_related('seasons')

    def index(self, request, **kwargs):
        return self.generic_list(request, self._competitions,
                                 templates=self.template_path('index.html'),
                                 paginate_by=self._competitions.count(),
                                 extra_context=kwargs)

    @competition_slug
    def competition(self, request, competition, extra_context, **kwargs):
        templates = self.template_path('competition.html', competition.slug)
        return self.generic_detail(request, self._competitions,
                                   slug=competition.slug,
                                   templates=templates,
                                   extra_context=extra_context)

    @competition_slug
    def season(self, request, competition, season, extra_context, **kwargs):
        templates = self.template_path(
            'season.html', competition.slug, season.slug)
        extra_context.update(season.matches.exclude(is_bye=True).aggregate(
            timeslot_count=Count('datetime', distinct=True),
            match_count=Count('pk', distinct=True),
            points_scored=Sum(
                Case(
                    When(
                        statistics__match__stage__division__season=season,
                        then=F('statistics__points')))),
            caps_awarded=Sum(
                Case(
                    When(
                        statistics__match__stage__division__season=season,
                        then=F('statistics__played')))),
        ))
        return self.generic_detail(request, competition.seasons,
                                   slug=season.slug,
                                   templates=templates,
                                   extra_context=extra_context)

    @competition_slug
    def season_videos(self, request, competition, season, extra_context, **kwargs):
        templates = self.template_path(
            'season_videos.html', competition.slug, season.slug)
        return self.generic_detail(request, competition.seasons,
                                   slug=season.slug,
                                   templates=templates,
                                   extra_context=extra_context)

    @competition_slug
    def club(self, request, competition, season, club, extra_context,
             **kwargs):
        templates = self.template_path(
            'club.html', competition.slug, season.slug, club.slug)
        return self.generic_detail(request, season.clubs,
                                   slug=club.slug,
                                   templates=templates,
                                   extra_context=extra_context)

    @competition_slug
    def division(self, request, competition, season, division, extra_context,
                 **kwargs):
        templates = self.template_path(
            'division.html', competition.slug, season.slug, division.slug)
        extra_context.update(division.matches.exclude(is_bye=True).aggregate(
            timeslot_count=Count('datetime', distinct=True),
            match_count=Count('pk', distinct=True),
            points_scored=Sum(
                Case(
                    When(
                        statistics__match__stage__division=division,
                        then=F('statistics__points')))),
            caps_awarded=Sum(
                Case(
                    When(
                        statistics__match__stage__division=division,
                        then=F('statistics__played')))),
        ))
        return self.generic_detail(request, season.divisions,
                                   slug=division.slug,
                                   templates=templates,
                                   extra_context=extra_context)

    @competition_slug
    def stage(self, request, competition, season, division, stage,
              extra_context, **kwargs):
        templates = self.template_path(
            'stage.html',
            competition.slug, season.slug, division.slug, stage.slug)
        extra_context['parent'] = stage
        extra_context.update(stage.matches.exclude(is_bye=True).aggregate(
            timeslot_count=Count('datetime', distinct=True),
            match_count=Count('pk', distinct=True),
            points_scored=Sum(
                Case(
                    When(
                        statistics__match__stage=stage,
                        then=F('statistics__points')))),
            caps_awarded=Sum(
                Case(
                    When(
                        statistics__match__stage=stage,
                        then=F('statistics__played')))),
        ))
        return self.generic_detail(request, division.stages,
                                   slug=stage.slug,
                                   templates=templates,
                                   extra_context=extra_context)

    @competition_slug
    def pool(self, request, competition, season, division, stage, pool,
             extra_context, **kwargs):
        # FIXME sadly the template name was not changed when we refactored
        #       pools to be subordinates of Stage.
        templates = self.template_path(
            'divisiongroup.html', competition.slug, season.slug, division.slug,
            stage.slug, pool.slug)
        extra_context.update(pool.matches.exclude(is_bye=True).aggregate(
            timeslot_count=Count('datetime', distinct=True),
            match_count=Count('pk', distinct=True),
            points_scored=Sum(
                Case(
                    When(
                        statistics__match__stage_group=pool,
                        then=F('statistics__points')))),
            caps_awarded=Sum(
                Case(
                    When(
                        statistics__match__stage_group=pool,
                        then=F('statistics__played')))),
        ))
        return self.generic_detail(request, stage.pools,
                                   slug=pool.slug,
                                   templates=templates,
                                   extra_context=extra_context)

    @competition_slug
    def team(self, request, competition, season, division, team, extra_context,
             **kwargs):
        templates = self.template_path(
            'team.html', competition.slug, season.slug, division.slug,
            team.slug)

        extra_context.update(team.matches.exclude(is_bye=True).aggregate(
            timeslot_count=Count('datetime', distinct=True),
            match_count=Count('pk', distinct=True),
            points_scored=Sum(
                Case(
                    When(
                        statistics__player__teamassociation__team=team,
                        then=F('statistics__points')))),
            caps_awarded=Sum(
                Case(
                    When(
                        statistics__player__teamassociation__team=team,
                        then=F('statistics__played')))),
        ))
        return self.generic_detail(request, division.teams,
                                   slug=team.slug,
                                   templates=templates,
                                   extra_context=extra_context)

    @competition_slug
    def calendar(self, request, season, club=None, division=None, team=None,
                 **kwargs):
        if team is not None:
            matches = team.matches
        elif division is not None:
            matches = division.matches
        elif club is not None:
            matches = club.matches.filter(stage__division__season=season)
        else:
            matches = season.matches

        # Do not include matches which have not had the time scheduled
        matches = matches.exclude(datetime__isnull=True)

        # Perform select_related to reduce extra queries
        matches = matches.select_related('stage__division__season')

        # For development server turn back plain text to make debugging easier
        if settings.DEBUG:
            content_type = 'text/plain'
        else:
            content_type = 'text/calendar'

        response = HttpResponse(content_type=content_type)

        cal = Calendar()
        cal.add('prodid', '-//Tournament Control//%s//' % request.get_host())
        cal.add('version', '2.0')

        for match in matches.order_by('datetime', 'play_at'):
            # Determine the resource path to the detailed match view
            path = self.reverse('match', kwargs={
                'match': match.pk,
                'division': match.stage.division.slug,
                'season': match.stage.division.season.slug,
                'competition': match.stage.division.season.competition.slug,
            })

            # Combine the resource path with our current request context
            url = request.build_absolute_uri(str(path))

            event = Event()
            event['uid'] = '%s@%s' % (
                base64.b64encode(smart_bytes(url)), request.get_host())
            event.add('summary', match.title)
            event.add('location', '{0} ({1})'.format(
                match.stage.division.title, match.stage.title))
            event.add('description', url)
            event.add('dtstart', match.datetime)
            # FIXME match duration should not be hardcoded
            event.add('dtend', match.datetime + timedelta(minutes=45))
            # FIXME should be the last modified time of the match
            event.add('dtstamp', timezone.now())

            cal.add_component(event)

        response.write(cal.to_ical())
        return response

    @competition_slug
    def match(self, request, competition, season, division, match,
              extra_context, **kwargs):
        if match.is_bye or match.home_team is None or match.away_team is None:
            # Temporary, realistically nobody should ever have this URL but we
            # will use GONE so that Google removes it from their index.
            return HttpResponseGone()
        templates = self.template_path(
            'match.html', competition.slug, season.slug, division.slug)
        return self.generic_detail(request, division.matches,
                                   pk=match.pk,
                                   templates=templates,
                                   extra_context=extra_context)

    @competition_slug
    def match_video(self, request, competition, season, division, match,
                    extra_context, **kwargs):
        if match.is_bye or match.home_team is None or match.away_team is None:
            # Temporary, realistically nobody should ever have this URL but we
            # will use GONE so that Google removes it from their index.
            return HttpResponseGone()
        templates = self.template_path(
            'videos.html', competition.slug, season.slug, division.slug)
        return self.generic_detail(request, division.matches,
                                   pk=match.pk,
                                   templates=templates,
                                   extra_context=extra_context)

        # TODO add related model to record videos
        # return self.generic_list(request, match.videos,
        #                          templates=templates,
        #                          extra_context=extra_context)

    @login_required_m
    @competition_slug
    def forfeit_list(self, request, competition, season, extra_context,
                     **kwargs):
        """
        List the matches that the visitor is permitted to forfeit. Must only
        show matches for teams that they are a registered member of (either as
        a player or other role).
        """
        try:
            teams = request.user.person.teams
        except Person.DoesNotExist:
            raise Http404

        matches = season.matches.filter(
            Q(home_team=teams) | Q(away_team=teams))

        context = {
            'matches': matches,
        }
        context.update(extra_context)

        templates = self.template_path('forfeit_list.html')
        return self.render(request, templates, context)

    @login_required_m
    @competition_slug
    def forfeit(self, request, competition, season, extra_context, match,
                **kwargs):
        """
        View that will allow a team member to forfeit a match they are due to
        be playing in.
        """
        try:
            teams = request.user.person.teams
        except Person.DoesNotExist:
            raise Http404

        matches = season.matches.filter(
            Q(home_team=teams) | Q(away_team=teams))

        redirect = self.redirect(self.reverse('forfeit-list', kwargs={
            'competition': competition.slug,
            'season': season.slug,
        }))

        if match not in matches:
            messages.add_message(
                request, messages.ERROR,
                "Attempting to forfeit a match you are not involved in.")
            return redirect

        # decide which team to forfeit
        if match.home_team in teams and match.away_team in teams:
            messages.add_message(
                request, messages.WARNING,
                _("You are a member of both teams for this match, please "
                  "contact the competition manager."))
            return redirect

        elif match.home_team in teams:
            team = match.home_team
        elif match.away_team in teams:
            team = match.away_team
        else:
            messages.add_message(
                request, messages.ERROR,
                _("You are not a member of either team, please contact the "
                  "competition manager."))
            return redirect

        if request.method == 'POST':
            # use the forfeit API to advise who made lodged the forfeit
            UNABLE_TO_POST = _("Unable to post forfeit, please contact the "
                               "competition manager.")
            try:
                success = match.forfeit(team, request.user.person)
            except AssertionError:
                # logger.exception("Failed precondition checks")
                messages.add_message(request, messages.ERROR, UNABLE_TO_POST)

            if success:
                messages.add_message(
                    request, messages.INFO,
                    _("The match has been forfeit, notifications will be sent "
                      "to affected parties shortly."))
            else:
                messages.add_message(request, messages.ERROR, UNABLE_TO_POST)

            return redirect

        context = {
            'match': match,
            'team': team,
        }
        context.update(extra_context)

        templates = self.template_path('forfeit.html')
        return self.render(request, templates, context)


class MultiCompetitionSite(CompetitionSite):

    kwargs_form_class = MultiConfigurationForm

    @classmethod
    def verbose_name(cls):
        return ugettext("Multiple Competitions")

    def __init__(self, name='competition', app_name='competition', **kwargs):
        self.node = kwargs.get('node')  # store the node for future reference
        super(CompetitionSite, self).__init__(name=name, app_name=app_name, **kwargs)
        self._competitions = Competition.objects.filter(enabled=True)
        if 'competition' in kwargs:
            q = reduce(or_, [Q(slug=slug) for slug in kwargs['competition']])
            self._competitions = self._competitions.filter(q)

    def get_urls(self):
        urlpatterns = [
            url(r'^$', self.index, name='index'),
            url(r'^(?P<competition>[^/]+)/', include(self.competition_urls())),
        ]
        return urlpatterns


class RegistrationSite(Application):

    def __init__(self, name='rego', app_name='rego', **kwargs):
        super(RegistrationSite, self).__init__(
            name=name, app_name=app_name, **kwargs)

    def get_urls(self):
        return []


class TournamentCalculatorSite(Application):

    def __init__(self, name='calculator', app_name='calculator', **kwargs):
        self.form_class = import_string(kwargs.pop(
            'form_class',
            'tournamentcontrol.competition.forms.TournamentScheduleForm'))
        super(TournamentCalculatorSite, self).__init__(
            name=name, app_name=app_name, **kwargs)

    def get_urls(self):
        live_form_class = import_string('tournamentcontrol.competition.forms.'
                                        'DivisionTournamentScheduleForm')
        urlpatterns = [
            url(r'^$', self.index,
                name='index',
                kwargs={
                    'form_class': self.form_class,
                }),
            url(r'^live/$', self.index,
                name='division',
                kwargs={
                    'form_class': live_form_class,
                    'template_name': 'live.html',
                }),
        ]
        return urlpatterns

    def index(self, request, form_class, template_name='index.html',
              **extra_context):
        context = {}

        if request.GET:
            form = form_class(data=request.GET)
        else:
            form = form_class()

        if form.is_valid():
            context.setdefault('data', form.get_context())

        context.setdefault('form', form)
        context.update(extra_context)

        templates = self.template_path(template_name)
        return self.render(request, templates, context)


class RankingSite(Application):

    kwargs_form_class = RankingConfigurationForm

    def __init__(self, name='ranking', app_name='ranking', **kwargs):
        super(RankingSite, self).__init__(
            name=name, app_name=app_name, **kwargs)

    def get_urls(self):
        urlpatterns = [
            url(r'^$', rank.IndexView.as_view(), name='index'),
            url(r'^(?P<year>\d{4})/', include([
                url(r'^$', rank.YearView.as_view(), name='year'),
                url(r'^(?P<month>[a-z]{3})/', include([
                    url(r'^$', rank.MonthView.as_view(), name='month'),
                    url(r'^(?P<day>\d{1,2})/', include([
                        url(r'^$', rank.DayView.as_view(), name='day'),
                        url(r'^(?P<slug>[^/]+)/', include([
                            url(r'^$', rank.DivisionView.as_view(), name='rank'),
                            url(r'^(?P<team>[^/]+)/$', rank.TeamView.as_view(), name='team'),
                        ])),
                    ])),
                ])),
            ])),
        ]
        return urlpatterns


competition = CompetitionSite()
registration = RegistrationSite()
calculator = TournamentCalculatorSite()
ranking = RankingSite()



# -*- coding: utf-8 -*-
# Generated by Django 1.10.4 on 2017-02-03 06:19
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('competition', '0023_competition__add__match_uuid_finalize'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='competition',
            options={'ordering': ('order',)},
        ),
        migrations.AlterModelOptions(
            name='division',
            options={'ordering': ('order',)},
        ),
        migrations.AlterModelOptions(
            name='season',
            options={'ordering': ('order',)},
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('competition', '0004_alter_person_gender'),
    ]

    operations = [
        migrations.AddField(
            model_name='club',
            name='facebook',
            field=models.URLField(max_length=255, blank=True),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='club',
            name='primary_position',
            field=models.CharField(help_text=b'Position of the primary contact', max_length=200, verbose_name=b'Position', blank=True),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='club',
            name='youtube',
            field=models.URLField(max_length=255, blank=True),
            preserve_default=True,
        ),
    ]



from __future__ import unicode_literals

import logging
import os.path

import django
from django.conf import settings
from django.conf.urls import url
from django.contrib import messages
from django.contrib.auth import forms, get_user_model, views
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.contrib.postgres.search import SearchVector
from django.core.paginator import EmptyPage, Paginator
from django.db import transaction
from django.forms.models import inlineformset_factory, modelformset_factory
from django.http import HttpResponse
from django.shortcuts import get_object_or_404, redirect
from django.template import RequestContext
from django.template.loader import render_to_string
from django.template.response import TemplateResponse
from django.urls import NoReverseMatch, reverse, reverse_lazy
from django.utils import timezone
from django.utils.encoding import smart_str
from django.utils.http import urlencode
from django.utils.module_loading import import_string
from django.utils.six.moves.urllib.parse import urljoin
from django.utils.translation import ugettext_lazy as _
from django.views.decorators.cache import patch_cache_control
from modelforms.forms import ModelForm
from six.moves import xrange
from touchtechnology.common.decorators import (
    login_required_m, node2extracontext, require_POST_m,
)
from touchtechnology.common.default_settings import PAGINATE_BY
from touchtechnology.common.utils import (
    get_403_or_None, get_all_perms_for_model_cached, get_objects_for_user,
    get_perms_for_model, model_and_manager, select_template_name,
)

logger = logging.getLogger(__name__)


class AlreadyRegistered(Exception):
    pass


class Application(object):

    kwargs_form_class = None
    model_form_bases = (ModelForm,)

    def __init__(self, name, app_name, node=None, **kwargs):
        self.name = name
        self.app_name = app_name
        self.node = node
        # store any instantiation keyword arguments to change behaviour
        self.kwargs = kwargs
        self._spawned = timezone.now()

    def __repr__(self):
        return '<{0}: {1}/{2}/{3}?{4}>'.format(self.__class__.__name__,
                                               self.name, self.app_name,
                                               self.node or '',
                                               urlencode(self.kwargs))

    def _get_namespace(self, cls=None):
        if cls is None:
            cls = self.__class__
        ns, __ = cls.__module__.split('.', 1)
        return ns if ns not in ('apps', self.app_name) else ''

    namespace = property(_get_namespace)

    @property
    def template_base(self):
        # configure the applications base template directory based on
        # namespace and app name
        return os.path.join(self.namespace, self.app_name)

    @property
    def urls(self):
        return self.get_urls(), self.app_name, self.name

    @classmethod
    def verbose_name(cls):
        return cls().app_name.title()

    @property
    def version(self):
        """
        Identify version number of the Application. If it is defined as a
        module we look up the import path until we find the module level
        ``get_version`` function, to a maximum depth of 4 levels.
        """
        ver = ''
        for depth in xrange(1, 5):
            parts = self.__module__.rsplit('.', depth)
            mod = parts[0]
            try:
                ver = import_string("%s.__version__" % mod)
            except ImportError:
                logger.warn('"__version__" not found at "%s"', mod)
            else:
                logger.debug('Identified version "%s" for "%s"', ver, mod)
                break
        name = self.verbose_name
        if callable(name):
            name = name()
        return '{name} {ver}'.format(name=_(name), ver=ver)

    @property
    def current_app(self):
        return self.name

    def reverse(self, name, args=None, kwargs=None, prefix=None):
        if args is None:
            args = ()
        if kwargs is None:
            kwargs = {}
        named_url = '%s:%s' % (self.app_name, name)
        if prefix:
            named_url = '%s:%s' % (prefix, named_url)
        for key in self.kwargs:
            kwargs.pop(key, None)
        return reverse_lazy(named_url, args=args, kwargs=kwargs,
                            current_app=self.current_app)

    def redirect(self, *args, **kwargs):
        "Thin wrapper around django.shortcuts.redirect to save on imports"
        return redirect(*args, **kwargs)

    def _template_path(self, filename, *args):
        args = list(reversed([a for a in args if a]))
        for index, arg in enumerate(args, 1):
            yield os.path.join(
                self.template_base, os.path.join(*args[:index]), filename)
        yield os.path.join(self.template_base, filename)

    def template_path(self, *args, **kwargs):
        return list(self._template_path(*args, **kwargs))

    def generic_permissions(self, request, model,
                            pk=None, instance=None,
                            post_save_redirect=None,
                            templates=None,
                            staff_only=None,
                            max_checkboxes=None,
                            **extra_context):
        # avoid circular import
        from touchtechnology.common.forms.auth import permissionformset_factory

        if instance is None:
            assert pk is not None
            instance = get_object_or_404(model, pk=pk)

        perms = get_perms_for_model(model, add=True)
        has_permission = get_403_or_None(request, perms, instance,
                                         return_403=True,
                                         accept_global_perms=True,
                                         any_perm=False)

        if has_permission and not request.user.is_superuser:
            return has_permission

        extra_context.update({
            'model': model,
        })

        content_type = ContentType.objects.get_for_model(model)
        queryset = Permission.objects.filter(content_type=content_type) \
                                     .exclude(codename__startswith='add_')
        formset_class = permissionformset_factory(
            model, staff_only=staff_only, max_checkboxes=max_checkboxes)

        if templates is None:
            templates = self.template_path(
                'permissions.html', model._meta.model_name)

        return self.generic_edit_multiple(
            request, model,
            queryset=queryset,
            formset_class=formset_class,
            formset_kwargs={
                'instance': instance,
                'queryset': queryset,
            },
            post_save_redirect=post_save_redirect,
            templates=templates,
            extra_context=extra_context)

    def generic_list(self, request, model_or_manager,
                     extra_context=None,
                     object_list_name=None,
                     paginate_by=None,
                     queryset=None,
                     templates=None,
                     list_templates=None,
                     perms=None,
                     permission_required=False,
                     accept_global_perms=True,
                     return_403=None,
                     search=None,
                     *args, **kwargs):

        model, manager = model_and_manager(model_or_manager)

        # If not explicitly declared, redirect the user to login if they are
        # currently anonymous. If they are already identified, we can safely
        # return a HTTP 403 Forbidden.
        if return_403 is None:
            return_403 = not request.user.is_anonymous

        if queryset is None:
            queryset = manager.all()

        # Search should be an iterable of fields to look into.
        if search is not None:
            terms = request.GET.get('search')
            if terms is not None:
                queryset = queryset.annotate(
                    search=SearchVector(*search),
                ).filter(search=terms)

        # Implement permission checking for list of objects, will filter the
        # list to those objects the user is entitled to work on if we do not
        # throw a HTTP 403 for no permission.
        if permission_required:
            if perms is None:
                perms = get_all_perms_for_model_cached(model)

            # Determine the user's permission to see the list using the
            # get_403_or_None - saves decorating view method with
            # @permission_required_or_403
            has_permission = get_403_or_None(
                request, perms, return_403=return_403,
                accept_global_perms=accept_global_perms, any_perm=True)

            global_perms = any([request.user.has_perm(p) for p in perms])

            # If the user does not have any global permissions then adjust the
            # queryset to return only the objects they can act on.
            if not global_perms:
                queryset = get_objects_for_user(
                    request.user, perms, queryset,
                    use_groups=True, any_perm=True)

            # If permission is denied, return the response. May already have
            # thrown an exception by now.
            if not queryset and has_permission is not None:
                return has_permission

        if extra_context is None:
            extra_context = {}

        if object_list_name is None:
            object_list_name = '{0}_list'.format(model._meta.model_name)

        if templates is None:
            templates = self.template_path('list.html',
                                           model._meta.model_name)

        if list_templates is None:
            list_templates = self.template_path('list.inc.html',
                                                model._meta.model_name)
            # Always tack the mvp theme version on the end as a fallback.
            list_templates.append('mvp/list.inc.html')

        # Determine the namespace of this Application so we can construct view
        # names to pass to the template engine to reverse CRUD URI's.
        create_url_name = '%s:add' % request.resolver_match.namespace
        create_url_kwargs = request.resolver_match.kwargs.copy()
        create_url_kwargs.pop('admin', None)
        create_url_kwargs.pop('component', None)

        # We can't leave this until the template is rendered because it throws
        # exceptions (stupid!) so we'll avoid lazy evaluation and pass None if
        # the create view doesn't exist for this model.
        try:
            create_url = reverse(create_url_name,
                                 args=request.resolver_match.args,
                                 kwargs=create_url_kwargs)
        except NoReverseMatch:
            create_url = None

        context = {
            'queryset': queryset,
            object_list_name: queryset,
            'object_list': queryset,
            'is_paginated': False,
            'page': None,
            'page_num': None,
            'paginator': None,
            'model': manager.none(),
            'template': select_template_name(templates),
            'list_template': select_template_name(list_templates),

            # Is search enabled?
            'searchable': bool(search),
            'search': request.GET.get('search', ''),

            # Pass to template the name permissions so we can re-use template
            # code to generically list and add/change/delete objects
            'add_perm': '%s.add_%s' % (model._meta.app_label, model._meta.model_name),
            'view_perm': '%s.view_%s' % (model._meta.app_label, model._meta.model_name),
            'change_perm': '%s.change_%s' % (model._meta.app_label, model._meta.model_name),
            'delete_perm': '%s.delete_%s' % (model._meta.app_label, model._meta.model_name),

            'create_url': create_url,
        }

        if paginate_by is None:
            paginate_by = PAGINATE_BY

        # Ensure that paginate_by is an integer to prevent ZeroDivisionError
        # inside the Django paginator steps.
        paginate_by = int(paginate_by)

        if paginate_by > 0:
            try:
                page_num = int(request.GET.get('page', 1))
            except TypeError:
                page_num = 1

            paginator = Paginator(queryset, paginate_by)

            try:
                page = paginator.page(page_num)
            except EmptyPage:
                page = paginator.page(paginator.num_pages)

            context.update({
                object_list_name: page.object_list,
                'object_list': page.object_list,
                'is_paginated': paginator.num_pages > 1,
                'page': page,
                'page_num': page_num,
                'paginator': paginator,
            })

        context.update(extra_context)
        return self.render(request, templates, context, *args, **kwargs)

    def generic_detail(self, request, model_or_manager,
                       templates=None,
                       extra_context=None,
                       **kwargs):
        model, manager = model_and_manager(model_or_manager)
        instance = get_object_or_404(manager, **kwargs)

        context = {
            'object': instance,
            model._meta.model_name: instance,
            'model': manager.none(),
        }
        context.update(extra_context or {})

        if templates is None:
            templates = self.template_path('detail.html',
                                           model._meta.model_name)

        return self.render(request, templates, context)

    def generic_edit(self, request, model_or_manager,
                     pk=None,
                     instance=None,
                     form_class=None,
                     form_kwargs=None,
                     form_fields='__all__',
                     form_widgets=None,
                     post_save_callback=lambda o: None,
                     post_save_redirect=None,
                     templates=None,
                     changed_messages=(
                         (messages.SUCCESS, _("Your {model} has been "
                                              "saved.")),
                     ),
                     unchanged_messages=(
                         (messages.INFO, _("Your {model} was not changed.")),
                     ),
                     permission_required=False,
                     perms=None,
                     accept_global_perms=True,
                     return_403=None,
                     related=None,
                     extra_context=None):

        model, manager = model_and_manager(model_or_manager)

        # If not explicitly declared, redirect the user to login if they are
        # currently anonymous. If they are already identified, we can safely
        # return a HTTP 403 Forbidden.
        if return_403 is None:
            return_403 = not request.user.is_anonymous

        if instance is None and pk is not None:
            instance = get_object_or_404(manager, pk=pk)

        if form_kwargs is None:
            form_kwargs = {}

        if extra_context is None:
            extra_context = {}

        if post_save_redirect is None:
            redirect_to = urljoin(request.path, '..') \
                if pk is None else request.path
            post_save_redirect = self.redirect(redirect_to)

        # Implement permission checking for specified object instance, throw a
        # HTTP 403 (using django-guardian configuration) when the user is not
        # entitled to edit the instance.
        if permission_required:

            # If no perms are specified, build sensible default using built in
            # permission types that come batteries included with Django.
            if perms is None:
                perms = get_perms_for_model(model, change=True)

                # When there is no pk value, we're doing a creation and should
                # have permission to create the object.
                if pk is None or instance.pk is None:
                    perms = get_perms_for_model(model, add=True)

            # Determine the user's permission to edit this object using the
            # get_403_or_None - saves decorating view method with
            # @permission_required_or_403
            has_permission = get_403_or_None(
                request, perms, obj=instance, return_403=return_403,
                accept_global_perms=accept_global_perms)

            # If permission is denied, return the response. May already have
            # thrown an exception by now.
            if has_permission is not None:
                return has_permission

        # If the developer has not provided a custom form, then dynamically
        # construct a default ModelForm for them.
        if form_class is None:
            meta_class = type(smart_str('Meta'), (), {'model': model,
                                        'fields': form_fields,
                                        'widgets': form_widgets})
            form_class = type(smart_str('EditForm'), self.model_form_bases,
                              {'Meta': meta_class})

        # Whether we've dynamically constructed our form_class or not, check to
        # ensure that we've inherited from all the bases. Log when we haven't,
        # but don't raise any exceptions.
        for base in self.model_form_bases:
            if not issubclass(form_class, base):
                logger.error('"%s" does not inherit "%s"', form_class, base)

        # Pass the instance to the form constructor if this is a ModelForm
        # subclass, otherwise it will need to be explicitly added to
        # form_kwargs if expected.
        if issubclass(form_class, ModelForm) and instance is not None:
            form_kwargs.setdefault('instance', instance)

        # Vanilla form processing here, take the post data and files, create
        # an instance of form_class, save and redirect.
        if request.method == "POST":
            form = form_class(data=request.POST, files=request.FILES,
                              **form_kwargs)
            if form.is_valid():
                replace = dict(
                    model=smart_str(model._meta.verbose_name),
                    models=smart_str(model._meta.verbose_name_plural),
                )
                if form.has_changed():
                    obj = form.save()
                    callback_res = post_save_callback(obj)
                    for level, message in changed_messages:
                        messages.add_message(request, level,
                                             message.format(**replace))
                    if callback_res is not None:
                        return callback_res
                else:
                    for level, message in unchanged_messages:
                        messages.add_message(request, level,
                                             message.format(**replace))
                return post_save_redirect
        else:
            form = form_class(**form_kwargs)

        if templates is None:
            templates = self.template_path('edit.html',
                                           model._meta.model_name)
            templates.append('edit.html')

        context = {
            'model': manager.none(),
            'template': select_template_name(templates),
            'form': form,
            'object': instance,
            'related': related,
            'cancel_url': post_save_redirect.url,
        }
        context.update(extra_context or {})

        return self.render(request, templates, context)

    def generic_edit_multiple(self, request, model_or_manager,
                              extra_context=None,
                              formset_class=None,
                              formset_kwargs=None,
                              formset_fields='__all__',
                              post_save_callback=lambda o: None,
                              post_save_redirect_args=None,
                              post_save_redirect_kwargs=None,
                              post_save_redirect=None,
                              templates=None,
                              changed_messages=None,
                              perms=None,
                              permission_required=False,
                              accept_global_perms=True,
                              return_403=None,
                              *args, **kwargs):

        if extra_context is None:
            extra_context = {}

        if formset_kwargs is None:
            formset_kwargs = {}

        if post_save_redirect_args is None:
            post_save_redirect_args = ()

        if post_save_redirect_kwargs is None:
            post_save_redirect_kwargs = {}

        if changed_messages is None:
            changed_messages = (
                (messages.SUCCESS, _("Your {models} have been saved.")),
            )

        # If not explicitly declared, redirect the user to login if they are
        # currently anonymous. If they are already identified, we can safely
        # return a HTTP 403 Forbidden.
        if return_403 is None:
            return_403 = not request.user.is_anonymous

        # Try and introspect the model if possible.
        model, manager = model_and_manager(model_or_manager)
        queryset = manager.all()

        # Implement permission checking for list of objects, will filter the
        # list to those objects the user is entitled to work on if we do not
        # throw a HTTP 403 for no permission.
        if permission_required:
            if perms is None:
                perms = get_all_perms_for_model_cached(model)

            # Determine the user's permission to see the list using the
            # get_403_or_None - saves decorating view method with
            # @permission_required_or_403
            has_permission = get_403_or_None(
                request, perms, return_403=return_403,
                accept_global_perms=accept_global_perms, any_perm=True)

            # If the user does not have any global permissions then adjust the
            # queryset to return only the objects they can act on.
            if not any([request.user.has_perm(p) for p in perms]):
                queryset = get_objects_for_user(
                    request.user, perms, queryset,
                    use_groups=True, any_perm=True)

            # If permission is denied, return the response. May already have
            # thrown an exception by now.
            if not queryset and has_permission is not None:
                return has_permission

            queryset = get_objects_for_user(
                request.user, perms, queryset, any_perm=True)

        # If queryset is not already specified as a keyword argument to the
        # formset attach it now. Value in formset_kwargs will take precedence.
        queryset = formset_kwargs.pop('queryset', queryset)

        # If the developer has not provided a custom formset, then dynamically
        # construct a default ModelFormSet for them.
        if formset_class is None:
            formset_class = modelformset_factory(model, extra=0,
                                                 fields=formset_fields)

        if post_save_redirect is None:
            redirect_to = urljoin(request.path, '..')
            post_save_redirect = self.redirect(redirect_to)

        # Vanilla formset processing here, take the post data and files, create
        # an instance of formset_class, save and redirect.
        if request.method == "POST":
            formset = formset_class(data=request.POST, files=request.FILES,
                                    queryset=queryset, **formset_kwargs)
            if formset.is_valid():
                replace = dict(
                    model=smart_str(model._meta.verbose_name),
                    models=smart_str(model._meta.verbose_name_plural),
                )
                objects = formset.save()
                map(post_save_callback, objects)
                for level, message in changed_messages:
                    messages.add_message(request, level,
                                         message.format(**replace))
                if isinstance(post_save_redirect, HttpResponse):
                    return post_save_redirect
                redirect_to = self.reverse(post_save_redirect,
                                           args=post_save_redirect_args,
                                           kwargs=post_save_redirect_kwargs)
                return redirect(redirect_to)
        else:
            formset = formset_class(queryset=queryset, **formset_kwargs)

        context = {
            'model': manager.none(),
            'formset': formset,
            'object_list': queryset,
        }
        context.update(extra_context)

        if templates is None:
            templates = self.template_path('edit_multiple.html',
                                           model._meta.model_name)
            templates.append('edit_multiple.html')

        return self.render(request, templates, context)

    def generic_edit_related(self, request, model_or_manager, related_model,
                             extra_context={},
                             form_class=None,
                             form_kwargs={},
                             form_fields='__all__',
                             form_widgets=None,
                             formset_class=None,
                             formset_kwargs={},
                             formset_fields='__all__',
                             instance=None,
                             pk=None,
                             post_save_callback=lambda o: None,
                             post_save_redirect_args=(),
                             post_save_redirect_kwargs={},
                             post_save_redirect='index',
                             post_save_related_callback=lambda o: None,
                             templates=None):

        model, manager = model_and_manager(model_or_manager)

        if instance is None and pk is not None:
            instance = get_object_or_404(manager, pk=pk)

        # If the developer has not provided a custom form, then dynamically
        # construct a default ModelForm for them.
        if form_class is None:
            meta_class = type(smart_str('Meta'), (), {'model': model, 'fields': form_fields, 'widgets': form_widgets})
            form_class = type(smart_str('EditForm'), self.model_form_bases, {'Meta': meta_class})

        # If the developer has not provided a custom formset, then dynamically
        # construct a default one with inlineformset_factory.
        if formset_class is None:
            formset_class = inlineformset_factory(
                model, related_model, fields=formset_fields, extra=0)

        # Add form and formset prefixes if not already present in the kwargs
        form_kwargs.setdefault('prefix', 'form')
        formset_kwargs.setdefault('prefix', 'formset')

        # Keep verbose names for model and related_model for use in messages
        model_verbose_name = model._meta.verbose_name.lower()
        related_verbose_name_plural = \
            related_model._meta.verbose_name_plural.lower()
        verbose = {
            'model_verbose_name': model_verbose_name,
            'related_verbose_name_plural': related_verbose_name_plural,
        }

        # Vanilla form processing here, take the post data and files, create
        # an instance of form_class, save and redirect.
        if request.method == "POST":
            sid = transaction.savepoint()

            form = form_class(data=request.POST, files=request.FILES,
                              instance=instance, **form_kwargs)

            if form.is_valid():
                obj = form.save()
                logger.debug('Saved %r to database.', obj)

                # Rebuild formset with updated instance object.
                formset = formset_class(data=request.POST,
                                        files=request.FILES,
                                        instance=obj, **formset_kwargs)

                if formset.is_valid():
                    logger.debug("Related formset is valid, if we don't "
                                 "complete successfully then an exception "
                                 "will bubble up.")
                    related = formset.save()
                    post_save_callback(obj)
                    map(post_save_related_callback, related)
                    transaction.savepoint_commit(sid)
                    messages.add_message(
                        request, messages.SUCCESS,
                        _("Your %(model_verbose_name)s and "
                          "%(related_verbose_name_plural)s "
                          "have been saved.") % verbose)
                    if isinstance(post_save_redirect, HttpResponse):
                        return post_save_redirect
                    redirect_to = self.reverse(
                        post_save_redirect,
                        args=post_save_redirect_args,
                        kwargs=post_save_redirect_kwargs)
                    return redirect(redirect_to)

                # if we've made it out without returning a HttpRedirect then
                # we should not be persisting the result.
                logger.debug("ROLLBACK, we were not redirected.")
                transaction.savepoint_rollback(sid)

            else:
                formset = formset_class(
                    data=request.POST, files=request.FILES,
                    instance=instance, **formset_kwargs)

        else:
            form = form_class(instance=instance, **form_kwargs)
            formset = formset_class(instance=instance, **formset_kwargs)

        context = {
            'model': manager.none(),
            'form': form,
            'formset': formset,
            'object': instance,
            'media': form.media + formset.media,
        }
        context.update(extra_context)

        if templates is None:
            templates = self.template_path('edit.related.html',
                                           model._meta.model_name)

        return self.render(request, templates, context)

    @require_POST_m
    def generic_delete(self, request, model_or_manager, pk,
                       post_delete_redirect=None,
                       perms=None,
                       permission_required=False,
                       accept_global_perms=True,
                       return_403=None,
                       **kwargs):

        model, manager = model_and_manager(model_or_manager)

        instance = get_object_or_404(manager, pk=pk)

        # If not explicitly declared, redirect the user to login if they are
        # currently anonymous. If they are already identified, we can safely
        # return a HTTP 403 Forbidden.
        if return_403 is None:
            return_403 = not request.user.is_anonymous

        # Implement permission checking for specified object instance, throw a
        # HTTP 403 (using django-guardian configuration) when the user is not
        # entitled to edit the instance.
        if permission_required:

            # If no perms are specified, build sensible default using built in
            # permission types that come batteries included with Django.
            if perms is None:
                perms = get_perms_for_model(model, delete=True)

            # Determine the user's permission to edit this object using the
            # get_403_or_None - saves decorating view method with
            # @permission_required_or_403
            has_permission = get_403_or_None(
                request, perms, obj=instance, return_403=return_403,
                accept_global_perms=accept_global_perms)

            # If permission is denied, return the response. May already have
            # thrown an exception by now.
            if has_permission is not None:
                return has_permission

        if post_delete_redirect is None:
            redirect_to = urljoin(request.path, '../..')
            post_delete_redirect = self.redirect(redirect_to)

        instance.delete()
        messages.add_message(
            request, messages.SUCCESS,
            _("The %s has been deleted.") % model._meta.verbose_name)
        return post_delete_redirect

    def context_instance(self, request, **kwargs):
        return RequestContext(request, current_app=self.current_app, **kwargs)

    def render_to_string(self, request, templates, context, **kwargs):
        context_instance = self.context_instance(request)
        return render_to_string(templates, context, context_instance, **kwargs)

    def render(self, request, templates, context, tz=None, **kwargs):
        logger.debug('{app}.render: {request.path}'.format(
            app=self.__class__.__name__, request=request))
        context.update({'application': self})
        kwargs.setdefault('context', context)

        response = TemplateResponse(request, templates, **kwargs)
        response.current_app = self.current_app

        if tz is not None:
            with timezone.override(tz):
                response.render()

        if not request.user.is_anonymous:
            patch_cache_control(response, private=True)
        return response


class AccountsSite(Application):

    def __init__(self, name='accounts', app_name='accounts', *args, **kwargs):
        super(AccountsSite, self).__init__(name=name, app_name=app_name)
        self.profile_form_class = kwargs.pop(
            'profile_form_class',
            'touchtechnology.common.forms_lazy.ProfileForm')
        self.user_model = get_user_model()

    @property
    def set_password_form(self):
        class SetPasswordForm(forms.SetPasswordForm):
            def __init__(slf, user, *a, **kw):
                super(SetPasswordForm, slf).__init__(user, *a, **kw)
                if user:
                    slf.user = self.user_model.objects.get(pk=user.pk)
        return SetPasswordForm

    @property
    def password_change_form(self):
        class PasswordChangeForm(forms.PasswordChangeForm):
            def __init__(slf, user, *a, **kw):
                super(PasswordChangeForm, slf).__init__(user, *a, **kw)
                if user:
                    slf.user = self.user_model.objects.get(pk=user.pk)
        return PasswordChangeForm

    @property
    def template_base(self):
        return 'registration'

    def get_urls(self):
        urlpatterns = [
            url(r'^login/$', self.login, name='login'),
            url(r'^logout/$', self.logout, name='logout'),
            url(r'^password-change/$',
                self.password_change, name='password_change'),
            url(r'^password-change/done/$',
                self.password_change_done, name='password_change_done'),
            url(r'^password-reset/$',
                self.password_reset, name='password_reset'),
            url(r'^password-reset/done/$',
                self.password_reset_done, name='password_reset_done'),
            url(r'^reset/(?P<uidb64>[0-9A-Za-z_\-]+)/'
                r'(?P<token>[0-9A-Za-z]{1,13}-[0-9A-Za-z]{1,20})/$',
                self.password_reset_confirm, name='password_reset_confirm'),
            url(r'^reset/done/$',
                self.password_reset_complete, name='password_reset_complete'),
            url(r'^profile/$', self.profile, name='profile'),
            url(r'^$', self.index, name='index'),
        ]
        return urlpatterns

    @node2extracontext
    def login(self, request, *args, **kwargs):
        form_class = getattr(
            settings, 'AUTHENTICATION_FORM_CLASS',
            'django.contrib.auth.forms.AuthenticationForm')
        authentication_form = import_string(form_class)
        if django.VERSION[0] < 2:
            kwargs.setdefault('current_app', self.name)
        return views.login(
            request, authentication_form=authentication_form, *args, **kwargs)

    @node2extracontext
    def logout(self, request, *args, **kwargs):
        if django.VERSION[0] < 2:
            kwargs.setdefault('current_app', self.name)
        return views.logout(request, *args, **kwargs)

    @node2extracontext
    def password_change(self, request, *args, **kwargs):
        post_change_redirect = kwargs.pop(
            'post_change_redirect', self.reverse('password_change_done'))
        password_change_form = kwargs.pop(
            'password_change_form', self.password_change_form)
        return views.password_change(
            request, post_change_redirect=post_change_redirect,
            password_change_form=password_change_form, *args, **kwargs)

    @node2extracontext
    def password_change_done(self, request, *args, **kwargs):
        return views.password_change_done(request, *args, **kwargs)

    @node2extracontext
    def password_reset(self, request, *args, **kwargs):
        post_reset_redirect = kwargs.pop(
            'post_reset_redirect', self.reverse('password_reset_done'))
        return views.password_reset(
            request, post_reset_redirect=post_reset_redirect, *args, **kwargs)

    @node2extracontext
    def password_reset_done(self, request, *args, **kwargs):
        return views.password_reset_done(request, *args, **kwargs)

    @node2extracontext
    def password_reset_confirm(self, request, *args, **kwargs):
        post_reset_redirect = kwargs.pop(
            'post_reset_redirect', self.reverse('password_reset_complete'))
        set_password_form = kwargs.pop(
            'set_password_form', self.set_password_form)
        return views.password_reset_confirm(
            request, post_reset_redirect=post_reset_redirect,
            set_password_form=set_password_form, *args, **kwargs)

    @node2extracontext
    def password_reset_complete(self, request, *args, **kwargs):
        return views.password_reset_complete(request, *args, **kwargs)

    def activation_complete(self, request, *args, **kwargs):
        context = {}
        context.update(kwargs)
        templates = self.template_path('activation_complete.html')
        return self.render(request, templates, context)

    def index(self, request, *args, **kwargs):
        return redirect(settings.LOGIN_REDIRECT_URL)

    @login_required_m
    def profile(self, request, *args, **kwargs):
        form_class = import_string(self.profile_form_class)

        if request.method == 'POST':
            form = form_class(data=request.POST, instance=request.user)
            if form.is_valid():
                if form.has_changed():
                    form.save()
                    messages.add_message(
                        request, messages.SUCCESS,
                        _("Your changes have been saved."))
                else:
                    messages.add_message(
                        request, messages.INFO,
                        _("You did not make any changes."))
                return redirect('.')
        else:
            form = form_class(instance=request.user)

        context = {
            'form': form,
        }
        context.update(kwargs)

        templates = self.template_path('profile.html')
        return self.render(request, templates, context)



from django.apps import AppConfig, apps


class ContentConfig(AppConfig):
    name = 'touchtechnology.content'

    def ready(self):
        from touchtechnology.content.utils import install_placeholder
        for config in apps.get_app_configs():
            install_placeholder(config.name)



from os.path import dirname, join

from django.apps import apps
from django.contrib.auth.models import Permission
from test_plus.test import TestCase
from touchtechnology.common.models import SitemapNode
from touchtechnology.common.tests.factories import UserFactory
from touchtechnology.content.models import Placeholder
from touchtechnology.content.tests import factories


class AdminTests(TestCase):

    def setUp(self):
        self.superuser = UserFactory.create(is_staff=True, is_superuser=True)
        self.staff = UserFactory.create(is_staff=True)
        self.staff.user_permissions.set(
            Permission.objects.filter(codename__endswith='_sitemapnode'))

        # Make sure the application config has been loaded for this app so as
        # to initialize the Placeholder instances.
        apps.get_app_config('content').ready()

        placeholder = Placeholder.objects.get(namespace='context')
        self.root_node = SitemapNode.objects.create(
            title='Simple Test',
            slug='home',
            object=placeholder,
            kwargs={'key': 'value'},
        )

    def get_text(self, filename):
        with open(join(dirname(__file__), filename)) as fp:
            text = fp.read()
        return text

    def test_sitemap_index_superuser(self):
        self.assertLoginRequired('admin:content:index')
        with self.login(self.superuser):
            self.get('admin:content:index')
            html = """
                <a role="button" href="/admin/content/application/{}/">
                    <i class="fa fa-pencil fa-fw"></i>
                    Edit
                </a>
            """.format(self.root_node.pk)
            self.assertResponseContains(html)

    def test_sitemap_index_staff_user(self):
        self.assertLoginRequired('admin:content:index')
        with self.login(self.staff):
            self.get('admin:content:index')
            html = "<span>Simple Test</span>"
            self.assertResponseContains(html)

    def test_edit_page(self):
        with self.login(self.superuser):
            self.assertGoodView('admin:content:page:add')

            copy = """<p>Sample page data.</p>"""
            data = {
                'parent-form-parent': '',
                'parent-form-title': 'Sample',
                'parent-form-short_title': '',
                'parent-form-enabled': '1',
                'parent-form-hidden_from_navigation': '0',
                'parent-form-hidden_from_sitemap': '0',
                'parent-form-slug': '',
                'parent-form-slug_locked': '0',

                'form-template': '',
                'form-keywords': '',
                'form-description': '',

                'formset-TOTAL_FORMS': '1',
                'formset-INITIAL_FORMS': '0',
                'formset-MIN_NUM_FORMS': '0',
                'formset-MAX_NUM_FORMS': '1000',
                'formset-0-copy': copy,
                'formset-0-label': 'copy',
                'formset-0-sequence': '1',
                'formset-0-id': '',
                'formset-0-page': '',
            }
            self.post('admin:content:page:add', data=data)
            self.response_302()

            node = SitemapNode.objects.latest('pk')
            self.assertEqual(node.get_absolute_url(), '/sample/')

            self.assertEqual(
                node.object.content.latest('pk').copy,
                copy)

            # Unfortunately the SitemapNodeMiddleware does not appear to be
            # having an effect?
            # self.get(node.get_absolute_url())
            # self.response_200()


class GoodViewTests(TestCase):

    def setUp(self):
        self.staff = UserFactory.create(is_staff=True, is_superuser=True)
        self.regular = UserFactory.create()

    def test_redirect_list(self):
        factories.RedirectFactory.create()

        with self.login(self.staff):
            self.assertGoodView('admin:content:redirect:list')
            self.response_200()



import argparse

from dateutil.parser import parse as date_parse
from django.core.management import BaseCommand
from django.utils.module_loading import import_string


def decay_function(path):
    try:
        return import_string(path)
    except ImportError:
        msg = 'Invalid decay function "{}" specified'.format(path)
        raise argparse.ArgumentTypeError(msg)


class Command(BaseCommand):

    def add_arguments(self, parser):
        parser.add_argument('--decay', default='tournamentcontrol.competition.rank.no_decay', type=decay_function)
        parser.add_argument('--start', type=date_parse)
        parser.add_argument('--at', type=date_parse)
        parser.add_argument('--json', type=argparse.FileType('wb'))

    def handle(self, start, at, decay, json, *args, **kwargs):
        from tournamentcontrol.competition.rank import rank, json_rank

        if not json:
            rank(start=start, at=at, decay=decay)
        else:
            json_rank(stream=json, start=start, at=at, decay=decay)



import uuid

from django.db import models
from touchtechnology.common.forms.fields import (
    EmailField as EmailFormField, GoogleMapsField, HTMLField as HTMLFormField,
    ModelChoiceField, ModelMultipleChoiceField, SelectDateField,
    SelectDateTimeField, SelectTimeField, SitemapNodeModelChoiceField,
    TemplatePathFormField,
)


class DateField(models.DateField):
    def formfield(self, form_class=SelectDateField, **kwargs):
        return super(DateField, self).formfield(
            form_class=form_class, **kwargs)


class EmailField(models.EmailField):
    """
    Custom EmailField which defaults to a length of 254 to be compliant with
    RFCs 3696 and 5321. Also will transform email address to lowercase unless
    advised otherwise as this SHOULD NOT be significant.
    """
    def __init__(self, max_length=254, lowercase=True, *args, **kwargs):
        super(EmailField, self).__init__(max_length=max_length,
                                         *args, **kwargs)
        self.lowercase = lowercase

    def clean(self, value, model_instance):
        value = super(EmailField, self).clean(value, model_instance)
        if self.lowercase:
            value = value.lower()
        return value

    def formfield(self, form_class=EmailFormField, **kwargs):
        return super(EmailField, self).formfield(
            form_class=form_class, lowercase=self.lowercase, **kwargs)


class TimeField(models.TimeField):
    def formfield(self, form_class=SelectTimeField, **kwargs):
        return super(TimeField, self).formfield(
            form_class=form_class, **kwargs)


class DateTimeField(models.DateTimeField):
    def formfield(self, form_class=SelectDateTimeField, **kwargs):
        return super(DateTimeField, self).formfield(
            form_class=form_class, **kwargs)

    def to_python(self, value):
        # See #13. When turning the database serialized value back into
        # a Python datetime if the value is '' or null an exception was
        # being raised. Explicitly check for these situations.
        if not value and (self.blank or self.null):
            return None
        return super(DateTimeField, self).to_python(value)


class LocationField(models.CharField):
    def formfield(self, form_class=GoogleMapsField, **kwargs):
        return super(LocationField, self).formfield(
            form_class=form_class, **kwargs)


class HTMLField(models.TextField):
    """
    A large string field for HTML content. It uses a custom widget in
    forms, and converts unicode characters to XML entities.
    """
    def formfield(self, form_class=HTMLFormField, widget=HTMLFormField.widget,
                  **kwargs):
        return super(HTMLField, self).formfield(
            form_class=form_class, widget=widget, **kwargs)


class ForeignKey(models.ForeignKey):
    def __init__(self, to, label_from_instance='name', *args, **kwargs):
        super(ForeignKey, self).__init__(to, *args, **kwargs)
        self.label_from_instance = label_from_instance

    def formfield(self, form_class=ModelChoiceField, **kwargs):
        kwargs.setdefault('label_from_instance', self.label_from_instance)
        return super(ForeignKey, self).formfield(
            form_class=form_class, **kwargs)


class ManyToManyField(models.ManyToManyField):
    """
    Custom `ManyToManyField` which overrides the formfield to use the
    much nicer `CheckboxSelectMultiple` widget instead of a multiple
    select list box.
    """
    def __init__(self, to, label_from_instance=None, *args, **kwargs):
        """
        Custom `__init__` because I don't want the 'Hold down "Control"...'
        message since I'm using the `CheckboxSelectMultiple` widget.
        """
        super(ManyToManyField, self).__init__(to, *args, **kwargs)
        self.label_from_instance = label_from_instance

    def formfield(self, form_class=ModelMultipleChoiceField, **kwargs):
        kwargs.setdefault('label_from_instance', self.label_from_instance)
        return super(ManyToManyField, self).formfield(
            form_class=form_class, **kwargs)


class NodeOneToOneField(models.OneToOneField):
    """
    Use when your OneToOneField has it's title on the SitemapNode.

    Purely aesthetic - gives us nicer labels in the default django admin.
    """
    def formfield(self, form_class=SitemapNodeModelChoiceField, **kwargs):
        return super(NodeOneToOneField, self).formfield(
            form_class=form_class, **kwargs)


class NodeForeignKey(models.ForeignKey):
    """
    Use when your ForeignKey has it's title on the SitemapNode.

    Purely aesthetic - gives us nicer labels in the default django admin.
    """
    def formfield(self, form_class=SitemapNodeModelChoiceField, **kwargs):
        return super(NodeForeignKey, self).formfield(
            form_class=form_class, **kwargs)


class TemplatePathField(models.CharField):
    """
    Field heavily based on `FilePathField` but designed to be able to accept
    custom templates from either the filesystem (as per `FilePathField`) or
    from the database by making use of our localised version of
    `django-dbtemplates`.
    """
    def __init__(self, verbose_name=None, name=None, template_base='',
                 template_folder='', match=None, recursive=False,
                 *args, **kwargs):
        self.template_base = template_base
        self.template_folder = template_folder
        self.match = match
        self.recursive = recursive
        super(TemplatePathField, self).__init__(verbose_name, name, **kwargs)

    def formfield(self, **kwargs):
        defaults = {
            'template_base': self.template_base,
            'template_folder': self.template_folder,
            'match': self.match,
            'recursive': self.recursive,
            'form_class': TemplatePathFormField,
        }
        defaults.update(kwargs)
        return super(TemplatePathField, self).formfield(**defaults)

    def get_internal_type(self):
        return "FilePathField"


class UUIDField(models.CharField):

    def __init__(self, verbose_name=None, name=None, auto=False, **kwargs):
        self.auto = auto
        # Fixed length, we're storing the UUIDs as text.
        kwargs['max_length'] = 36
        if auto:
            # Do not let the user edit UUIDs if they are auto assigned
            kwargs['editable'] = False
            kwargs['blank'] = True
        super(UUIDField, self).__init__(verbose_name, name, **kwargs)

    def pre_save(self, model_instance, add):
        value = super(UUIDField, self).pre_save(model_instance, add)
        if not value and self.auto:
            value = str(uuid.uuid1())
            setattr(model_instance, self.attname, value)
        return value
